"""Body model for DOCX documents."""

from __future__ import annotations

from typing import Any, Dict, Iterable, List, Tuple, Type

from .base import Models


class Body(Models):
    """Represents the main document container (body or textbox content)."""

    def __init__(self) -> None:
        super().__init__()
        self.sections: List[Any] = []

    # ------------------------------------------------------------------
    def _allowed_types(self) -> Tuple[Type[Models], ...]:
        from .paragraph import Paragraph
        from .table import Table
        from .image import Image
        from .textbox import TextBox

        return (Paragraph, Table, Image, TextBox)

    def _attach_child(self, model: Models) -> Models:
        self.add_child(model)
        return model

    def add_model(self, model: Models) -> Models:
        if not isinstance(model, Models):
            raise TypeError(f"Body can only contain Models instances, got {type(model)!r}")
        if not isinstance(model, self._allowed_types()):
            allowed = ", ".join(t.__name__ for t in self._allowed_types())
            raise TypeError(f"Unsupported model type {type(model).__name__}; allowed: {allowed}")
        return self._attach_child(model)

    def add_paragraph(self, paragraph: Models) -> Models:
        return self.add_model(paragraph)

    def add_table(self, table: Models) -> Models:
        return self.add_model(table)

    def add_image(self, image: Models) -> Models:
        return self.add_model(image)

    def add_textbox(self, textbox: Models) -> Models:
        return self.add_model(textbox)

    # ------------------------------------------------------------------
    def _filter_children(self, type_: Type[Models]) -> List[Models]:
        return [child for child in self.children if isinstance(child, type_)]

    def get_paragraphs(self) -> List[Models]:
        from .paragraph import Paragraph

        return self._filter_children(Paragraph)

    def get_tables(self) -> List[Models]:
        from .table import Table

        return self._filter_children(Table)

    def get_images(self) -> List[Models]:
        from .image import Image

        return self._filter_children(Image)

    def get_textboxes(self) -> List[Models]:
        from .textbox import TextBox

        return self._filter_children(TextBox)

    def _iter_text_sources(self) -> Iterable[str]:
        for child in self.children:
            if hasattr(child, "get_text"):
                text = child.get_text()
                if text:
                    yield text
            elif hasattr(child, "text"):
                text = getattr(child, "text")
                if text:
                    yield str(text)

    def get_text(self) -> str:
        return "\n".join(self._iter_text_sources()) if self.children else ""

    # ------------------------------------------------------------------
    def get_paragraphs_recursive(self) -> List[Models]:
        from .paragraph import Paragraph

        result: List[Models] = []
        for child in self.children:
            if isinstance(child, Paragraph):
                result.append(child)
            elif hasattr(child, "get_paragraphs_recursive"):
                result.extend(child.get_paragraphs_recursive())
            elif hasattr(child, "get_paragraphs"):
                result.extend(child.get_paragraphs())
        return result

    def get_tables_recursive(self) -> List[Models]:
        from .table import Table

        result: List[Models] = []
        for child in self.children:
            if isinstance(child, Table):
                result.append(child)
                if hasattr(child, "get_tables_recursive"):
                    result.extend(child.get_tables_recursive())
            elif hasattr(child, "get_tables_recursive"):
                result.extend(child.get_tables_recursive())
            elif hasattr(child, "get_tables"):
                result.extend(child.get_tables())
        return result

    def get_images_recursive(self) -> List[Models]:
        from .image import Image

        result: List[Models] = []
        for child in self.children:
            if isinstance(child, Image):
                result.append(child)
            elif hasattr(child, "get_images_recursive"):
                result.extend(child.get_images_recursive())
            elif hasattr(child, "get_images"):
                result.extend(child.get_images())
        return result
    
    def search(self, search_text: str, case_sensitive: bool = False, 
               search_in_tables: bool = True, search_in_textboxes: bool = True) -> List[Dict[str, Any]]:
        """
        Search for text in body content.
        
        Args:
            search_text: Text to search for
            case_sensitive: Whether search should be case sensitive
            search_in_tables: Whether to search in table content
            search_in_textboxes: Whether to search in textbox content
            
        Returns:
            List of search results with context information
        """
        import re
        
        results = []
        search_pattern = search_text if case_sensitive else search_text.lower()
        
        for i, child in enumerate(self.children):
            if hasattr(child, 'get_text'):
                text = child.get_text()
                if text:
                    search_text_content = text if case_sensitive else text.lower()
                    
                    # Find all occurrences
                    for match in re.finditer(re.escape(search_pattern), search_text_content):
                        results.append({
                            'element_index': i,
                            'element_type': type(child).__name__,
                            'element': child,
                            'match_start': match.start(),
                            'match_end': match.end(),
                            'context': text[max(0, match.start() - 50):match.end() + 50],
                            'full_text': text
                        })
            
            # Search in tables if enabled
            if search_in_tables and hasattr(child, 'get_tables_recursive'):
                for table in child.get_tables_recursive():
                    table_results = self._search_in_table(table, search_pattern, case_sensitive)
                    results.extend(table_results)
            
            # Search in textboxes if enabled
            if search_in_textboxes and hasattr(child, 'get_textboxes_recursive'):
                for textbox in child.get_textboxes_recursive():
                    textbox_results = self._search_in_textbox(textbox, search_pattern, case_sensitive)
                    results.extend(textbox_results)
        
        return results
    
    def _search_in_table(self, table, search_pattern: str, case_sensitive: bool) -> List[Dict[str, Any]]:
        """Search for text in table content."""
        results = []
        
        for row_idx, row in enumerate(table.rows):
            for cell_idx, cell in enumerate(row.cells):
                if hasattr(cell, 'get_text'):
                    text = cell.get_text()
                    if text:
                        search_text_content = text if case_sensitive else text.lower()
                        
                        for match in re.finditer(re.escape(search_pattern), search_text_content):
                            results.append({
                                'element_index': f"table_row_{row_idx}_cell_{cell_idx}",
                                'element_type': 'TableCell',
                                'element': cell,
                                'match_start': match.start(),
                                'match_end': match.end(),
                                'context': text[max(0, match.start() - 50):match.end() + 50],
                                'full_text': text,
                                'table_info': {
                                    'table': table,
                                    'row_index': row_idx,
                                    'cell_index': cell_idx
                                }
                            })
        
        return results
    
    def _search_in_textbox(self, textbox, search_pattern: str, case_sensitive: bool) -> List[Dict[str, Any]]:
        """Search for text in textbox content."""
        results = []
        
        if hasattr(textbox, 'get_text'):
            text = textbox.get_text()
            if text:
                search_text_content = text if case_sensitive else text.lower()
                
                for match in re.finditer(re.escape(search_pattern), search_text_content):
                    results.append({
                        'element_index': f"textbox_{id(textbox)}",
                        'element_type': 'TextBox',
                        'element': textbox,
                        'match_start': match.start(),
                        'match_end': match.end(),
                        'context': text[max(0, match.start() - 50):match.end() + 50],
                        'full_text': text
                    })
        
        return results
    
    def replace(self, search_text: str, replace_text: str, case_sensitive: bool = False,
                replace_in_tables: bool = True, replace_in_textboxes: bool = True) -> int:
        """
        Replace text in body content.
        
        Args:
            search_text: Text to search for
            replace_text: Text to replace with
            case_sensitive: Whether search should be case sensitive
            replace_in_tables: Whether to replace in table content
            replace_in_textboxes: Whether to replace in textbox content
            
        Returns:
            Number of replacements made
        """
        import re
        
        total_replacements = 0
        search_pattern = search_text if case_sensitive else search_text.lower()
        
        for child in self.children:
            if hasattr(child, 'get_text') and hasattr(child, 'set_text'):
                text = child.get_text()
                if text:
                    search_text_content = text if case_sensitive else text.lower()
                    
                    # Count occurrences
                    matches = list(re.finditer(re.escape(search_pattern), search_text_content))
                    if matches:
                        # Replace text
                        new_text = text
                        for match in reversed(matches):  # Replace from end to avoid index issues
                            new_text = new_text[:match.start()] + replace_text + new_text[match.end():]
                        
                        child.set_text(new_text)
                        total_replacements += len(matches)
            
            # Replace in tables if enabled
            if replace_in_tables and hasattr(child, 'get_tables_recursive'):
                for table in child.get_tables_recursive():
                    table_replacements = self._replace_in_table(table, search_pattern, replace_text, case_sensitive)
                    total_replacements += table_replacements
            
            # Replace in textboxes if enabled
            if replace_in_textboxes and hasattr(child, 'get_textboxes_recursive'):
                for textbox in child.get_textboxes_recursive():
                    textbox_replacements = self._replace_in_textbox(textbox, search_pattern, replace_text, case_sensitive)
                    total_replacements += textbox_replacements
        
        return total_replacements
    
    def _replace_in_table(self, table, search_pattern: str, replace_text: str, case_sensitive: bool) -> int:
        """Replace text in table content."""
        total_replacements = 0
        
        for row in table.rows:
            for cell in row.cells:
                if hasattr(cell, 'get_text') and hasattr(cell, 'set_text'):
                    text = cell.get_text()
                    if text:
                        search_text_content = text if case_sensitive else text.lower()
                        
                        # Count occurrences
                        matches = list(re.finditer(re.escape(search_pattern), search_text_content))
                        if matches:
                            # Replace text
                            new_text = text
                            for match in reversed(matches):  # Replace from end to avoid index issues
                                new_text = new_text[:match.start()] + replace_text + new_text[match.end():]
                            
                            cell.set_text(new_text)
                            total_replacements += len(matches)
        
        return total_replacements
    
    def _replace_in_textbox(self, textbox, search_pattern: str, replace_text: str, case_sensitive: bool) -> int:
        """Replace text in textbox content."""
        total_replacements = 0
        
        if hasattr(textbox, 'get_text') and hasattr(textbox, 'set_text'):
            text = textbox.get_text()
            if text:
                search_text_content = text if case_sensitive else text.lower()
                
                # Count occurrences
                matches = list(re.finditer(re.escape(search_pattern), search_text_content))
                if matches:
                    # Replace text
                    new_text = text
                    for match in reversed(matches):  # Replace from end to avoid index issues
                        new_text = new_text[:match.start()] + replace_text + new_text[match.end():]
                    
                    textbox.set_text(new_text)
                    total_replacements += len(matches)
        
        return total_replacements
    
    def find_by_style(self, style_name: str) -> List[Any]:
        """
        Find elements by style name.
        
        Args:
            style_name: Name of the style to search for
            
        Returns:
            List of elements with the specified style
        """
        elements = []
        
        for child in self.children:
            if hasattr(child, 'style') and child.style == style_name:
                elements.append(child)
            
            # Search in tables
            if hasattr(child, 'get_tables_recursive'):
                for table in child.get_tables_recursive():
                    for row in table.rows:
                        for cell in row.cells:
                            if hasattr(cell, 'style') and cell.style == style_name:
                                elements.append(cell)
            
            # Search in textboxes
            if hasattr(child, 'get_textboxes_recursive'):
                for textbox in child.get_textboxes_recursive():
                    if hasattr(textbox, 'style') and textbox.style == style_name:
                        elements.append(textbox)
        
        return elements


"""
Entry point for running DocQuill as a module.

Usage:
    python -m docquill input.docx --output output.html --format html
    docquill input.docx --output output.pdf --format pdf
"""

from .cli import main

if __name__ == "__main__":
    main()



"""Page engine for creating pages with pre-calculated headers, footers, and margins.

This engine handles:
- Pre-calculation of header/footer heights
- Different header/footer variants (first page, odd/even pages)
- Page creation with already calculated margins
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from .unified_layout import LayoutPage
from .geometry import Margins, Size


@dataclass(slots=True)
class PageConfig:
    """Configuration for page creation."""
    page_size: Size
    base_margins: Margins
    header_height: float = 0.0
    footer_height: float = 0.0
    header_height_first: Optional[float] = None  # Different header for first page
    footer_height_first: Optional[float] = None  # Different footer for first page
    header_height_even: Optional[float] = None  # Different header for even pages
    footer_height_even: Optional[float] = None  # Different footer for even pages
    
    @property
    def content_height(self) -> float:
        """Calculate available content height (excluding header/footer)."""
        return self.page_size.height - self.base_margins.top - self.base_margins.bottom - self.header_height - self.footer_height
    
    @property
    def content_top(self) -> float:
        """Top of content area in PDF coordinates."""
        return self.page_size.height - self.base_margins.top - self.header_height
    
    @property
    def content_bottom(self) -> float:
        """Bottom of content area in PDF coordinates."""
        return self.base_margins.bottom + self.footer_height
    
    def get_header_height(self, page_number: int) -> float:
        """Get header height for specific page number.
        
        Args:
            page_number: Page number (1-based)
            
        Returns:
            Header height for this page
        """
        if page_number == 1 and self.header_height_first is not None:
            return self.header_height_first
        if page_number % 2 == 0 and self.header_height_even is not None:
            return self.header_height_even
        # Odd pages (except first) use default header_height
        return self.header_height
    
    def get_footer_height(self, page_number: int) -> float:
        """Get footer height for specific page number.
        
        Args:
            page_number: Page number (1-based)
            
        Returns:
            Footer height for this page
        """
        if page_number == 1 and self.footer_height_first is not None:
            return self.footer_height_first
        if page_number % 2 == 0 and self.footer_height_even is not None:
            return self.footer_height_even
        return self.footer_height
    
    def get_content_height(self, page_number: int) -> float:
        """Get content height for specific page number.
        
        Args:
            page_number: Page number (1-based)
            
        Returns:
            Available content height for this page
        """
        header_h = self.get_header_height(page_number)
        footer_h = self.get_footer_height(page_number)
        return self.page_size.height - self.base_margins.top - self.base_margins.bottom - header_h - footer_h
    
    def get_content_top(self, page_number: int) -> float:
        """Get content top position for specific page number.
        
        Args:
            page_number: Page number (1-based)
            
        Returns:
            Top of content area in PDF coordinates
        """
        header_h = self.get_header_height(page_number)
        return self.page_size.height - self.base_margins.top - header_h
    
    def get_content_bottom(self, page_number: int) -> float:
        """Get content bottom position for specific page number.
        
        Args:
            page_number: Page number (1-based)
            
        Returns:
            Bottom of content area in PDF coordinates
        """
        footer_h = self.get_footer_height(page_number)
        return self.base_margins.bottom + footer_h


class PageEngine:
    """Engine for creating pages with pre-calculated headers, footers, and margins."""
    
    def __init__(self, config: PageConfig):
        """Initialize page engine.
        
        Args:
            config: Page configuration with pre-calculated header/footer heights
        """
        self.config = config
        self.current_page_number = 1
    
    def create_page(self, page_number: Optional[int] = None) -> LayoutPage:
        """Create a new page with pre-calculated margins.
        
        Args:
            page_number: Page number (if None, uses current page number)
            
        Returns:
            LayoutPage with pre-calculated margins
        """
        if page_number is None:
            page_number = self.current_page_number
        else:
            self.current_page_number = page_number
        
        # Get header/footer heights for this page
        header_height = self.config.get_header_height(page_number)
        footer_height = self.config.get_footer_height(page_number)
        
        # Calculate effective margins (base margins + header/footer)
        effective_margins = Margins(
            left=self.config.base_margins.left,
            right=self.config.base_margins.right,
            top=self.config.base_margins.top + header_height,
            bottom=self.config.base_margins.bottom + footer_height,
        )
        
        # Create page with pre-calculated header/footer heights
        page = LayoutPage(
            number=page_number,
            size=self.config.page_size,
            margins=effective_margins,
        )
        page.header_height = header_height
        page.footer_height = footer_height
        
        return page
    
    def starting_cursor(self, page_number: Optional[int] = None) -> float:
        """Get starting cursor position for a page.
        
        Args:
            page_number: Page number (if None, uses current page number)
            
        Returns:
            Top Y position of content area in PDF coordinates
        """
        if page_number is None:
            page_number = self.current_page_number
        return self.config.get_content_top(page_number)
    
    def remaining_space(self, y_cursor: float, page_number: Optional[int] = None) -> float:
        """Compute remaining vertical space above the bottom margin.
        
        Args:
            y_cursor: Current Y position in PDF coordinates
            page_number: Page number (if None, uses current page number)
            
        Returns:
            Remaining space in points
        """
        if page_number is None:
            page_number = self.current_page_number
        content_bottom = self.config.get_content_bottom(page_number)
        return max(0.0, y_cursor - content_bottom)
    
    def fits_on_page(self, y_cursor: float, block_height: float, page_number: Optional[int] = None) -> bool:
        """Check whether the block can fit on the current page.
        
        Args:
            y_cursor: Top Y position of block in PDF coordinates
            block_height: Height of block in points
            page_number: Page number (if None, uses current page number)
            
        Returns:
            True if block fits, False otherwise
        """
        return self.remaining_space(y_cursor, page_number) >= block_height
    
    def next_page(self) -> int:
        """Increment page number and return new page number.
        
        Returns:
            New page number
        """
        self.current_page_number += 1
        return self.current_page_number
    
    def reset(self) -> None:
        """Reset page engine for new document."""
        self.current_page_number = 1




"""Run model for DOCX documents."""

from typing import List, Dict, Any, Optional, Tuple, Type
from .base import Models

class Run(Models):
    """Represents a run of text with consistent formatting."""
    
    def __init__(
        self,
        text="",
        style=None,
        space="default",
        has_break=False,
        has_tab=False,
        has_drawing=False,
        from_sdt=False,
        break_type: Optional[str] = None,
    ):
        """Initialize run with text, style and space handling."""
        super().__init__()
        self.text: str = text
        self.style: Optional[Dict[str, Any]] = style or {}
        self.space: str = space  # "default" or "preserve"
        self.allowed_models: Tuple[Type[Models], ...] = (
            # Field, Hyperlink  # inline fragments - will be imported when needed
        )
        
        # Special elements
        self.has_break: bool = has_break
        self.has_tab: bool = has_tab
        self.has_drawing: bool = has_drawing
        self.from_sdt: bool = from_sdt  # Marks runs that came from SDT elements
        self.break_type: Optional[str] = break_type  # e.g. textWrapping, page, column, line
        
        # Drawing content
        self.textbox: Optional[List['Run']] = None  # Textbox content as list of Run objects
        self.image: Optional[Any] = None  # Image object
        
        # Formatting attributes
        self.bold: bool = False
        self.italic: bool = False
        self.underline: bool = False
        self.font_name: Optional[str] = None
        self.font_size: Optional[int] = None
        self.color: Optional[str] = None
        self.highlight: Optional[str] = None
        self.superscript: bool = False
        self.subscript: bool = False
        self.strike_through: bool = False
        # Advanced formatting attributes
        self.strikethrough: bool = False
        self.double_strikethrough: bool = False
        self.vertical_align: Optional[str] = None
        self.background: Optional[str] = None
        self.effect: Optional[Dict[str, Any]] = None
        self.outline: Optional[Dict[str, Any]] = None
        self.shadow: Optional[Dict[str, Any]] = None
        self.emboss: bool = False
        self.imprint: bool = False
        self.no_proof: bool = False
        self.web_hidden: bool = False
        
        # Footnote and endnote references
        self.footnote_refs: List[str] = []
        self.endnote_refs: List[str] = []
    
    def add_text(self, text: str):
        """Add text to run."""
        self.text += text
    
    def add_field(self, field):
        """Add field to run."""
        self.add_child(field)
    
    def add_hyperlink(self, hyperlink):
        """Add hyperlink to run."""
        self.add_child(hyperlink)
    
    def set_style(self, style):
        """Set run style."""
        self.style = style
    
    def get_text(self):
        """Return stored text respecting the run's xml:space behaviour."""
        # Always return text as-is to preserve spacing
        # Empty runs with space should return space, not empty string
        return self.text
    
    def is_bold(self):
        """Check if run is bold."""
        if isinstance(self.style, dict):
            return self.style.get('bold', False)
        return 'bold' in str(self.style).lower() if self.style else False
    
    def is_italic(self):
        """Check if run is italic."""
        if isinstance(self.style, dict):
            return self.style.get('italic', False)
        return 'italic' in str(self.style).lower() if self.style else False
    
    def is_underline(self):
        """Check if run is underlined."""
        if isinstance(self.style, dict):
            return self.style.get('underline', False)
        return 'underline' in str(self.style).lower() if self.style else False
    
    def set_bold(self, bold: bool):
        """Set bold formatting."""
        self.bold = bold
    
    def set_italic(self, italic: bool):
        """Set italic formatting."""
        self.italic = italic
    
    def set_underline(self, underline: bool):
        """Set underline formatting."""
        self.underline = underline
    
    def set_font_name(self, font_name: str):
        """Set font name."""
        self.font_name = font_name
    
    def set_font_size(self, font_size: int):
        """Set font size in points."""
        self.font_size = font_size
    
    def set_color(self, color: str):
        """Set text color."""
        self.color = color
    
    def set_highlight(self, highlight: str):
        """Set highlight color."""
        self.highlight = highlight
    
    def set_superscript(self, superscript: bool):
        """Set superscript formatting."""
        self.superscript = superscript
        if superscript:
            self.subscript = False
    
    def set_subscript(self, subscript: bool):
        """Set subscript formatting."""
        self.subscript = subscript
        if subscript:
            self.superscript = False
    
    def set_strike_through(self, strike_through: bool):
        """Set strike-through formatting."""
        self.strike_through = strike_through
    
    def apply_formatting(self, formatting: Dict[str, Any]):
        """Apply formatting from dictionary."""
        if 'bold' in formatting:
            self.set_bold(formatting['bold'])
        if 'italic' in formatting:
            self.set_italic(formatting['italic'])
        if 'underline' in formatting:
            self.set_underline(formatting['underline'])
        if 'font_name' in formatting:
            self.set_font_name(formatting['font_name'])
        if 'font_size' in formatting:
            self.set_font_size(formatting['font_size'])
        if 'color' in formatting:
            self.set_color(formatting['color'])
        if 'highlight' in formatting:
            self.set_highlight(formatting['highlight'])
        if 'superscript' in formatting:
            self.set_superscript(formatting['superscript'])
        if 'subscript' in formatting:
            self.set_subscript(formatting['subscript'])
        if 'strike_through' in formatting:
            self.set_strike_through(formatting['strike_through'])
    
    def get_formatting(self) -> Dict[str, Any]:
        """Get formatting as dictionary."""
        return {
            'bold': self.bold,
            'italic': self.italic,
            'underline': self.underline,
            'font_name': self.font_name,
            'font_size': self.font_size,
            'color': self.color,
            'highlight': self.highlight,
            'superscript': self.superscript,
            'subscript': self.subscript,
            'strike_through': self.strike_through
        }
    
    def clone(self):
        """Clone run with all formatting preserved."""
        cloned = super().clone()
        cloned.text = self.text
        cloned.style = self.style.copy() if self.style else None
        cloned.space = self.space
        cloned.bold = self.bold
        cloned.italic = self.italic
        cloned.underline = self.underline
        cloned.font_name = self.font_name
        cloned.font_size = self.font_size
        cloned.color = self.color
        cloned.highlight = self.highlight
        cloned.superscript = self.superscript
        cloned.subscript = self.subscript
        cloned.strike_through = self.strike_through
        cloned.break_type = self.break_type
        return cloned
    
    def to_html(self) -> str:
        """Convert run to HTML representation."""
        if not self.text:
            return ""
        
        html_text = self.text
        html_tags = []
        
        # Apply formatting tags
        if self.bold:
            html_tags.append(('b', '</b>'))
        if self.italic:
            html_tags.append(('i', '</i>'))
        if self.underline:
            html_tags.append(('u', '</u>'))
        if self.strike_through:
            html_tags.append(('s', '</s>'))
        if self.superscript:
            html_tags.append(('sup', '</sup>'))
        if self.subscript:
            html_tags.append(('sub', '</sub>'))
        
        # Apply style attributes
        style_parts = []
        if self.font_name:
            style_parts.append(f"font-family: {self.font_name}")
        if self.font_size:
            style_parts.append(f"font-size: {self.font_size}pt")
        if self.color:
            style_parts.append(f"color: {self.color}")
        if self.highlight:
            style_parts.append(f"background-color: {self.highlight}")
        
        # Build HTML
        if style_parts:
            style_attr = f' style="{"; ".join(style_parts)}"'
        else:
            style_attr = ""
        
        # Wrap with tags
        for open_tag, close_tag in html_tags:
            html_text = f"<{open_tag}>{html_text}</{open_tag}>"
        
        # Add span with style if needed
        if style_attr:
            html_text = f'<span{style_attr}>{html_text}</span>'
        
        return html_text



"""
Rust Canvas Wrapper - Canvas-like API for Rust PDF renderer.

This module provides a ReportLab Canvas-compatible API that delegates to Rust renderer.
All business logic stays in Python - Rust only handles low-level PDF operations.
"""

import logging
import time
from typing import Optional, Union, Tuple, List
from pathlib import Path

logger = logging.getLogger(__name__)

# Try to import Rust renderer
# First try the unified PyPI package, then fallback to legacy names
rust_pdf_canvas = None
HAS_RUST_RENDERER = False

try:
    # Unified PyPI package: pip install docquill-rust
    import docquill_rust as rust_pdf_canvas
    HAS_RUST_RENDERER = True
except ImportError:
    try:
        # Legacy PyPI package: pip install docquill-pdf-rust
        import docquill_pdf_rust as rust_pdf_canvas
        HAS_RUST_RENDERER = True
    except ImportError:
        try:
            # Development/legacy name
            import rust_pdf_canvas
            HAS_RUST_RENDERER = True
        except ImportError:
            pass


class RustCanvas:
    """
    Canvas-like wrapper for Rust PDF renderer.
    
    Provides the same API as ReportLab Canvas, but delegates to Rust renderer.
    This allows PDFCompiler to use Rust renderer without changing business logic.
    """
    
    def __init__(self, rust_renderer, page_width: float, page_height: float, metrics_engine=None, batch_size: int = 0):
        """
        Initialize Rust Canvas wrapper.
        
        Args:
            rust_renderer: Rust PdfRenderer instance (from pdf_renderer_rust)
            page_width: Page width in points
            page_height: Page height in points
            metrics_engine: Optional TextMetricsEngine for text width measurement
            batch_size: Maximum number of commands to batch before flushing (default: 500)
                        Set to 0 to disable batching (immediate execution)
        """
        if not HAS_RUST_RENDERER:
            raise ImportError("rust_pdf_canvas not available")
        
        self._renderer = rust_renderer
        self._page_width = page_width
        self._page_height = page_height
        self._metrics_engine = metrics_engine
        self._batch_size = batch_size
        self._batch_commands = []  # List of command dicts to batch
        
        # Performance tracking (lightweight - just counting, no timing overhead)
        self._call_count = 0
        self._rust_call_time = 0.0
        self._enable_profiling = False  # Set to True to enable detailed profiling (adds timing overhead)
        
        # Canvas state tracking (for compatibility with ReportLab)
        self._current_fill_color = (0.0, 0.0, 0.0)  # RGB
        self._current_stroke_color = (0.0, 0.0, 0.0)  # RGB
        self._current_line_width = 1.0
        self._current_font_name = "Helvetica"
        self._current_font_size = 12.0
        self._current_dash = None
    
    # Canvas operations (matching ReportLab API)
    
    def _call_rust(self, method_name: str, *args, **kwargs):
        """
        Helper method to call Rust renderer method with optional profiling.
        
        Args:
            method_name: Name of the method to call on self._renderer
            *args, **kwargs: Arguments to pass to the method
        """
        if self._enable_profiling:
            t0 = time.perf_counter()
            method = getattr(self._renderer, method_name)
            result = method(*args, **kwargs)
            elapsed = time.perf_counter() - t0
            self._call_count += 1
            self._rust_call_time += elapsed
            return result
        else:
            method = getattr(self._renderer, method_name)
            return method(*args, **kwargs)
    
    def get_profiling_stats(self) -> dict:
        """Get profiling statistics (call count and total time)."""
        return {
            "call_count": self._call_count,
            "rust_call_time": self._rust_call_time,
            "avg_call_time": self._rust_call_time / self._call_count if self._call_count > 0 else 0.0,
        }
    
    def reset_profiling(self):
        """Reset profiling statistics."""
        self._call_count = 0
        self._rust_call_time = 0.0
    
    def _add_command(self, cmd_dict: dict):
        """Add a command to the batch queue."""
        if self._batch_size == 0:
            # Batching disabled - execute immediately by calling Rust methods directly
            # This avoids dict creation/parsing overhead
            cmd_type = cmd_dict.get("type")
            try:
                # Count calls (lightweight, no timing overhead)
                self._call_count += 1
                
                if cmd_type == "SaveState":
                    self._renderer.canvas_save_state()
                elif cmd_type == "RestoreState":
                    self._renderer.canvas_restore_state()
                elif cmd_type == "SetFillColor":
                    self._renderer.canvas_set_fill_color(cmd_dict["r"], cmd_dict["g"], cmd_dict["b"])
                elif cmd_type == "SetStrokeColor":
                    self._renderer.canvas_set_stroke_color(cmd_dict["r"], cmd_dict["g"], cmd_dict["b"])
                elif cmd_type == "SetLineWidth":
                    self._renderer.canvas_set_line_width(cmd_dict["width"])
                elif cmd_type == "SetDash":
                    self._renderer.canvas_set_dash(cmd_dict["pattern"])
                elif cmd_type == "SetFont":
                    self._renderer.canvas_set_font(cmd_dict["name"], cmd_dict["size"])
                elif cmd_type == "SetOpacity":
                    self._renderer.canvas_set_opacity(cmd_dict["opacity"])
                elif cmd_type == "Rect":
                    self._renderer.canvas_rect(cmd_dict["x"], cmd_dict["y"], cmd_dict["width"], cmd_dict["height"], cmd_dict["fill"], cmd_dict["stroke"])
                elif cmd_type == "RoundRect":
                    self._renderer.canvas_round_rect(cmd_dict["x"], cmd_dict["y"], cmd_dict["width"], cmd_dict["height"], cmd_dict["radius"], cmd_dict["fill"], cmd_dict["stroke"])
                elif cmd_type == "Line":
                    self._renderer.canvas_line(cmd_dict["x1"], cmd_dict["y1"], cmd_dict["x2"], cmd_dict["y2"])
                elif cmd_type == "DrawString":
                    self._renderer.canvas_draw_string(cmd_dict["x"], cmd_dict["y"], cmd_dict["text"])
                elif cmd_type == "DrawImage":
                    self._renderer.canvas_draw_image(cmd_dict["x"], cmd_dict["y"], cmd_dict["width"], cmd_dict["height"], cmd_dict["image_data"])
                elif cmd_type == "Translate":
                    self._renderer.canvas_translate(cmd_dict["x"], cmd_dict["y"])
                elif cmd_type == "Rotate":
                    self._renderer.canvas_rotate(cmd_dict["angle"])
                elif cmd_type == "Scale":
                    self._renderer.canvas_scale(cmd_dict["x"], cmd_dict["y"])
                elif cmd_type == "Transform":
                    self._renderer.canvas_transform(cmd_dict["matrix"])
                else:
                    logger.warning(f"Unknown command type: {cmd_type}")
            except Exception as e:
                if not hasattr(self, '_error_logged'):
                    logger.error(f"Error executing command {cmd_type}: {type(e).__name__}: {e}", exc_info=True)
                    self._error_logged = True
                raise
        else:
            self._batch_commands.append(cmd_dict)
            # Auto-flush if batch is full
            if len(self._batch_commands) >= self._batch_size:
                self._flush_batch()
    
    def _execute_command_directly(self, cmd: dict):
        """Execute a single command directly without batching."""
        cmd_type = cmd.get("type")
        try:
            if cmd_type == "SaveState":
                self._renderer.canvas_save_state()
            elif cmd_type == "RestoreState":
                self._renderer.canvas_restore_state()
            elif cmd_type == "SetFillColor":
                self._renderer.canvas_set_fill_color(cmd["r"], cmd["g"], cmd["b"])
            elif cmd_type == "SetStrokeColor":
                self._renderer.canvas_set_stroke_color(cmd["r"], cmd["g"], cmd["b"])
            elif cmd_type == "SetLineWidth":
                self._renderer.canvas_set_line_width(cmd["width"])
            elif cmd_type == "SetDash":
                self._renderer.canvas_set_dash(cmd["pattern"])
            elif cmd_type == "SetFont":
                self._renderer.canvas_set_font(cmd["name"], cmd["size"])
            elif cmd_type == "SetOpacity":
                self._renderer.canvas_set_opacity(cmd["opacity"])
            elif cmd_type == "Rect":
                self._renderer.canvas_rect(cmd["x"], cmd["y"], cmd["width"], cmd["height"], cmd["fill"], cmd["stroke"])
            elif cmd_type == "RoundRect":
                self._renderer.canvas_round_rect(cmd["x"], cmd["y"], cmd["width"], cmd["height"], cmd["radius"], cmd["fill"], cmd["stroke"])
            elif cmd_type == "Line":
                self._renderer.canvas_line(cmd["x1"], cmd["y1"], cmd["x2"], cmd["y2"])
            elif cmd_type == "DrawString":
                self._renderer.canvas_draw_string(cmd["x"], cmd["y"], cmd["text"])
            elif cmd_type == "DrawImage":
                self._renderer.canvas_draw_image(cmd["x"], cmd["y"], cmd["width"], cmd["height"], cmd["image_data"])
            elif cmd_type == "Translate":
                self._renderer.canvas_translate(cmd["x"], cmd["y"])
            elif cmd_type == "Rotate":
                self._renderer.canvas_rotate(cmd["angle"])
            elif cmd_type == "Scale":
                self._renderer.canvas_scale(cmd["x"], cmd["y"])
            elif cmd_type == "Transform":
                self._renderer.canvas_transform(cmd["matrix"])
            else:
                logger.warning(f"Unknown command type: {cmd_type}")
        except Exception as e:
            # Log error but don't crash - this helps debug the issue
            if not hasattr(self, '_error_logged'):
                logger.error(f"Error executing command {cmd_type}: {type(e).__name__}: {e}", exc_info=True)
                self._error_logged = True
            raise
    
    def _flush_batch(self, commands: Optional[List[dict]] = None):
        """Flush batched commands to Rust renderer."""
        if commands is None:
            commands = self._batch_commands
            self._batch_commands = []
        
        if not commands:
            return
        
        # Call Rust batch API
        # Rust expects &PyList, which PyO3 will convert automatically from Python list
        try:
            # Check if method exists first
            if not hasattr(self._renderer, 'canvas_run_batch'):
                raise AttributeError("canvas_run_batch method not found")
            
            # Pass the list directly - PyO3 should convert Python list to &PyList automatically
            # Log first batch to verify it's being called
            if not hasattr(self, '_batch_first_call_logged'):
                logger.info(f"Using batch API: flushing {len(commands)} commands")
                self._batch_first_call_logged = True
            
            # Try to call batch API - PyO3 should handle conversion
            try:
                self._renderer.canvas_run_batch(commands)
            except TypeError as te:
                # If PyO3 can't convert automatically, try converting to PyList manually
                import pyo3
                import sys
                # This shouldn't be necessary, but if it fails, we'll use fallback
                raise te
        except (AttributeError, TypeError) as e:
            # Fallback: if batch API doesn't exist or has wrong signature, execute commands individually
            # Only log once to avoid spam
            if not hasattr(self, '_batch_fallback_logged'):
                logger.warning(f"canvas_run_batch not available or failed ({type(e).__name__}: {e}), falling back to individual calls")
                self._batch_fallback_logged = True
            for cmd in commands:
                cmd_type = cmd.get("type")
                if cmd_type == "SaveState":
                    self._renderer.canvas_save_state()
                elif cmd_type == "RestoreState":
                    self._renderer.canvas_restore_state()
                elif cmd_type == "SetFillColor":
                    self._renderer.canvas_set_fill_color(cmd["r"], cmd["g"], cmd["b"])
                elif cmd_type == "SetStrokeColor":
                    self._renderer.canvas_set_stroke_color(cmd["r"], cmd["g"], cmd["b"])
                elif cmd_type == "SetLineWidth":
                    self._renderer.canvas_set_line_width(cmd["width"])
                elif cmd_type == "SetDash":
                    self._renderer.canvas_set_dash(cmd["pattern"])
                elif cmd_type == "SetFont":
                    self._renderer.canvas_set_font(cmd["name"], cmd["size"])
                elif cmd_type == "SetOpacity":
                    self._renderer.canvas_set_opacity(cmd["opacity"])
                elif cmd_type == "Rect":
                    self._renderer.canvas_rect(cmd["x"], cmd["y"], cmd["width"], cmd["height"], cmd["fill"], cmd["stroke"])
                elif cmd_type == "RoundRect":
                    self._renderer.canvas_round_rect(cmd["x"], cmd["y"], cmd["width"], cmd["height"], cmd["radius"], cmd["fill"], cmd["stroke"])
                elif cmd_type == "Line":
                    self._renderer.canvas_line(cmd["x1"], cmd["y1"], cmd["x2"], cmd["y2"])
                elif cmd_type == "DrawString":
                    self._renderer.canvas_draw_string(cmd["x"], cmd["y"], cmd["text"])
                elif cmd_type == "DrawImage":
                    self._renderer.canvas_draw_image(cmd["x"], cmd["y"], cmd["width"], cmd["height"], cmd["image_data"])
                elif cmd_type == "Translate":
                    self._renderer.canvas_translate(cmd["x"], cmd["y"])
                elif cmd_type == "Rotate":
                    self._renderer.canvas_rotate(cmd["angle"])
                elif cmd_type == "Scale":
                    self._renderer.canvas_scale(cmd["x"], cmd["y"])
                elif cmd_type == "Transform":
                    self._renderer.canvas_transform(cmd["matrix"])
    
    def saveState(self):
        """Save current canvas state."""
        self._add_command({"type": "SaveState"})
    
    def restoreState(self):
        """Restore previous canvas state."""
        self._add_command({"type": "RestoreState"})
    
    def setFillColor(self, color):
        """
        Set fill color.
        
        Args:
            color: Color object or tuple (r, g, b) or hex string
        """
        rgb = self._parse_color(color)
        self._current_fill_color = rgb
        self._add_command({"type": "SetFillColor", "r": rgb[0], "g": rgb[1], "b": rgb[2]})
    
    def setStrokeColor(self, color):
        """
        Set stroke color.
        
        Args:
            color: Color object or tuple (r, g, b) or hex string
        """
        rgb = self._parse_color(color)
        self._current_stroke_color = rgb
        self._add_command({"type": "SetStrokeColor", "r": rgb[0], "g": rgb[1], "b": rgb[2]})
    
    def setLineWidth(self, width: float):
        """Set line width."""
        self._current_line_width = width
        self._add_command({"type": "SetLineWidth", "width": width})
    
    def setDash(self, *args):
        """
        Set dash pattern.
        
        Args:
            *args: Dash pattern (e.g., [6, 3] or 6, 3)
        """
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            pattern = list(args[0])
        else:
            pattern = list(args)
        self._current_dash = pattern
        self._add_command({"type": "SetDash", "pattern": pattern})
    
    def setFont(self, name: str, size: float):
        """Set font name and size."""
        self._current_font_name = name
        self._current_font_size = size
        self._add_command({"type": "SetFont", "name": name, "size": size})
    
    def setOpacity(self, opacity: float):
        """Set current drawing opacity (0.0 - 1.0)."""
        try:
            value = max(0.0, min(1.0, float(opacity)))
        except (TypeError, ValueError):
            value = 1.0
        self._add_command({"type": "SetOpacity", "opacity": value})
    
    def rect(self, x: float, y: float, width: float, height: float, fill: int = 0, stroke: int = 1):
        """
        Draw rectangle.
        
        Args:
            x, y: Position
            width, height: Dimensions
            fill: 1 to fill, 0 otherwise
            stroke: 1 to stroke, 0 otherwise
        """
        self._add_command({
            "type": "Rect",
            "x": x,
            "y": y,
            "width": width,
            "height": height,
            "fill": bool(fill),
            "stroke": bool(stroke),
        })
    
    def roundRect(self, x: float, y: float, width: float, height: float, radius: float, fill: int = 0, stroke: int = 1):
        """
        Draw rounded rectangle.
        
        Args:
            x, y: Position
            width, height: Dimensions
            radius: Corner radius
            fill: 1 to fill, 0 otherwise
            stroke: 1 to stroke, 0 otherwise
        """
        self._add_command({
            "type": "RoundRect",
            "x": x,
            "y": y,
            "width": width,
            "height": height,
            "radius": radius,
            "fill": bool(fill),
            "stroke": bool(stroke),
        })
    
    def line(self, x1: float, y1: float, x2: float, y2: float):
        """Draw line from (x1, y1) to (x2, y2)."""
        self._add_command({"type": "Line", "x1": x1, "y1": y1, "x2": x2, "y2": y2})
    
    def drawString(self, x: float, y: float, text: str):
        """
        Draw text string.
        
        Args:
            x, y: Position
            text: Text to draw
        """
        self._add_command({"type": "DrawString", "x": x, "y": y, "text": text})
    
    def drawImage(self, image_path: Union[str, Path], x: float, y: float, width: Optional[float] = None, height: Optional[float] = None, mask: Optional[str] = None, preserveAspectRatio: bool = False):
        """
        Draw image.
        
        Args:
            image_path: Path to image file
            x, y: Position
            width, height: Dimensions (optional)
            mask: Mask mode (ignored for now)
            preserveAspectRatio: Whether to preserve aspect ratio (ignored for now)
        """
        # Load image data
        image_path = Path(image_path)
        if not image_path.exists():
            logger.warning(f"Image not found: {image_path}")
            return
        
        with open(image_path, 'rb') as f:
            image_data = f.read()
        
        # Use provided dimensions or detect from image
        if width is None or height is None:
            # Try to detect dimensions (simplified - would need image library)
            if width is None:
                width = 100.0  # Default
            if height is None:
                height = 100.0  # Default
        
        self._add_command({
            "type": "DrawImage",
            "x": x,
            "y": y,
            "width": width,
            "height": height,
            "image_data": image_data,
        })
    
    def translate(self, x: float, y: float):
        """Translate coordinate system."""
        self._add_command({"type": "Translate", "x": x, "y": y})
    
    def rotate(self, angle: float):
        """Rotate coordinate system (degrees)."""
        # Convert degrees to radians (Rust expects radians)
        import math
        radians = math.radians(angle)
        self._add_command({"type": "Rotate", "angle": radians})
    
    def scale(self, x: float, y: float):
        """Scale coordinate system."""
        self._add_command({"type": "Scale", "x": x, "y": y})
    
    def transform(self, a: float, b: float, c: float, d: float, e: float, f: float):
        """Apply transformation matrix."""
        self._add_command({"type": "Transform", "matrix": [a, b, c, d, e, f]})
    
    def setPageSize(self, size: Tuple[float, float]):
        """Set page size."""
        width, height = size
        self._page_width = width
        self._page_height = height
        # Note: Page size is set when creating new page, this just updates tracking
        try:
            self._renderer.set_page_size(width, height)
        except AttributeError:
            # Method might not exist, just update local tracking
            pass
    
    def showPage(self):
        """Finish current page and start new one."""
        # Flush any pending batched commands before page change
        self._flush_batch()
        # This is handled by the renderer, not the canvas
        pass
    
    def flush(self):
        """Manually flush all batched commands."""
        self._flush_batch()
    
    def stringWidth(self, text: str, font_name: str, font_size: float) -> float:
        """
        Calculate text width.
        
        Args:
            text: Text to measure
            font_name: Font name
            font_size: Font size in points
            
        Returns:
            Width in points
        """
        # Use TextMetricsEngine if available (same as ReportLab Canvas)
        if self._metrics_engine is not None:
            try:
                style = {
                    "font_name": font_name,
                    "font_size": font_size,
                }
                result = self._metrics_engine.measure_text(text, style)
                return result.get("width", 0.0)
            except Exception:
                pass
        
        # Fallback: simple estimation based on character count
        # Average character width for most fonts is approximately 0.6 * font_size
        avg_char_width = font_size * 0.6
        return len(text) * avg_char_width
    
    # Helper methods
    
    def _parse_color(self, color) -> Tuple[float, float, float]:
        """
        Parse color to RGB tuple (0.0-1.0).
        
        Supports:
        - ReportLab Color objects
        - Tuples (r, g, b) in 0.0-1.0 range
        - Hex strings (#RRGGBB)
        """
        # ReportLab Color object
        if hasattr(color, 'red') and hasattr(color, 'green') and hasattr(color, 'blue'):
            return (color.red, color.green, color.blue)
        
        # Tuple or list
        if isinstance(color, (tuple, list)):
            if len(color) >= 3:
                # Normalize to 0.0-1.0 range
                r, g, b = color[0], color[1], color[2]
                if r > 1.0 or g > 1.0 or b > 1.0:
                    # Assume 0-255 range
                    return (r / 255.0, g / 255.0, b / 255.0)
                return (r, g, b)
        
        # Hex string
        if isinstance(color, str) and color.startswith('#'):
            color = color[1:]
            if len(color) == 6:
                r = int(color[0:2], 16) / 255.0
                g = int(color[2:4], 16) / 255.0
                b = int(color[4:6], 16) / 255.0
                return (r, g, b)
        
        # Default to black
        logger.warning(f"Could not parse color: {color}, using black")
        return (0.0, 0.0, 0.0)




"""

TextAlignmentEngine - calculating X position for text relative to column width.

Supports:
- left: left alignment (default)
- center: centering
- right: right alignment
- justify: justification (requires additional logic)

"""

from typing import Dict, Any
from .geometry import Rect


class TextAlignmentEngine:
    """
    Silnik do obliczania pozycji X dla tekstu na podstawie alignment.
    """
    
    @staticmethod
    def calculate_x(
        rect: Rect,
        text_width: float,
        alignment: str = "left"
    ) -> float:
        """

        Calculates X position for text based on alignment.

        Args:
        rect: Rect of text area
        text_width: Text width in points
        alignment: Alignment ("left", "center", "right", "justify")

        Returns:
        X position for text

        """
        alignment = alignment.lower() if alignment else "left"
        
        if alignment == "center":
            # Centering
            x = rect.x + (rect.width - text_width) / 2
            return max(rect.x, x)  # Don't go beyond left edge
        
        elif alignment == "right":
            # Right alignment
            x = rect.x + rect.width - text_width
            return max(rect.x, x)  # Don't go beyond left edge
        
        elif alignment == "justify":
            # Justification - text fills entire width
            # Return left edge, justification requires additional logic in renderer
            return rect.x
        
        else:  # "left" or default
            # Left alignment
            return rect.x
    
    @staticmethod
    def get_alignment_from_style(style: Dict[str, Any]) -> str:
        """

        Gets alignment from style.

        Args:
        style: Dictionary with styles

        Returns:
        Alignment string ("left", "center", "right", "justify")

        """
        alignment = style.get("alignment") or style.get("text_align") or style.get("align", "left")
        
        # Normalizuj
        alignment = str(alignment).lower()
        
        # Map different variants
        if alignment in ("left", "start", "l"):
            return "left"
        elif alignment in ("center", "middle", "c"):
            return "center"
        elif alignment in ("right", "end", "r"):
            return "right"
        elif alignment in ("justify", "justified", "j"):
            return "justify"
        else:
            return "left"  # Default
    
    @staticmethod
    def calculate_text_position(
        rect: Rect,
        text_width: float,
        style: Dict[str, Any]
    ) -> float:
        """

        Calculates X position for text based on rect, text width and style.

        Args:
        rect: Rect of text area
        text_width: Text width in points
        style: Dictionary with styles (contains alignment)

        Returns:
        X position for text

        """
        alignment = TextAlignmentEngine.get_alignment_from_style(style)
        return TextAlignmentEngine.calculate_x(rect, text_width, alignment)




"""
Footnote model for DOCX documents.

Handles footnote functionality, properties, content, validation, and numbering.
"""

from typing import Dict, Any, Optional, List, List
from .base import Models
import logging

logger = logging.getLogger(__name__)

class Footnote(Models):
    """
    Represents a footnote in the document.
    
    Handles footnote functionality, properties, content, and validation.
    """
    
    def __init__(self, footnote_id: str = "", footnote_type: str = "normal", 
                 content: str = "", properties: Dict[str, Any] = None):
        """
        Initialize footnote.
        
        Args:
            footnote_id: Footnote identifier
            footnote_type: Type of footnote (normal, separator, continuation)
            content: Footnote content
            properties: Footnote properties
        """
        super().__init__()
        self.footnote_id = footnote_id
        self.footnote_type = footnote_type
        self.content = content
        self.properties = properties or {}
        self.validation_errors = []
        
        logger.debug(f"Footnote initialized: {footnote_id}")
    
    def set_footnote_id(self, footnote_id: str) -> None:
        """
        Set footnote ID.
        
        Args:
            footnote_id: Footnote identifier
        """
        if not footnote_id or not isinstance(footnote_id, str):
            raise ValueError("Footnote ID must be a non-empty string")
        
        self.footnote_id = footnote_id
        logger.debug(f"Footnote ID set to: {footnote_id}")
    
    def set_footnote_type(self, footnote_type: str) -> None:
        """
        Set footnote type.
        
        Args:
            footnote_type: Type of footnote
        """
        if not footnote_type or not isinstance(footnote_type, str):
            raise ValueError("Footnote type must be a non-empty string")
        
        self.footnote_type = footnote_type
        logger.debug(f"Footnote type set to: {footnote_type}")
    
    def set_content(self, content: str) -> None:
        """
        Set footnote content.
        
        Args:
            content: Footnote content
        """
        if not isinstance(content, str):
            raise ValueError("Footnote content must be a string")
        
        self.content = content
        logger.debug(f"Footnote content set: {len(content)} characters")
    
    def get_footnote_id(self) -> str:
        """
        Get footnote ID.
        
        Returns:
            Footnote identifier
        """
        return self.footnote_id
    
    def get_content(self) -> str:
        """
        Get footnote content.
        
        Returns:
            Footnote content
        """
        return self.content
    
    def get_footnote_type(self) -> str:
        """
        Get footnote type.
        
        Returns:
            Footnote type
        """
        return self.footnote_type
    
    def set_property(self, key: str, value: Any) -> None:
        """
        Set footnote property.
        
        Args:
            key: Property key
            value: Property value
        """
        self.properties[key] = value
        logger.debug(f"Footnote property set: {key} = {value}")
    
    def get_property(self, key: str, default: Any = None) -> Any:
        """
        Get footnote property.
        
        Args:
            key: Property key
            default: Default value if property not found
            
        Returns:
            Property value or default
        """
        return self.properties.get(key, default)
    
    def validate(self) -> bool:
        """
        Validate footnote.
        
        Returns:
            True if footnote is valid, False otherwise
        """
        self.validation_errors = []
        
        # Validate footnote ID
        if not self.footnote_id:
            self.validation_errors.append("Footnote ID is required")
        
        # Validate footnote type
        if not self.footnote_type:
            self.validation_errors.append("Footnote type is required")
        
        # Validate content
        if not self.content:
            self.validation_errors.append("Footnote content is required")
        
        # Validate properties
        if not isinstance(self.properties, dict):
            self.validation_errors.append("Footnote properties must be a dictionary")
        
        is_valid = len(self.validation_errors) == 0
        logger.debug(f"Footnote validation: {'valid' if is_valid else 'invalid'}")
        
        return is_valid
    
    def get_validation_errors(self) -> List[str]:
        """
        Get validation errors.
        
        Returns:
            List of validation errors
        """
        return self.validation_errors.copy()
    
    def to_dict(self) -> Dict[str, Any]:
        """
        Convert footnote to dictionary.
        
        Returns:
            Dictionary representation of footnote
        """
        return {
            'type': 'footnote',
            'footnote_id': self.footnote_id,
            'footnote_type': self.footnote_type,
            'content': self.content,
            'properties': self.properties.copy(),
            'validation_errors': self.validation_errors.copy()
        }
    
    def from_dict(self, data: Dict[str, Any]) -> None:
        """
        Load footnote from dictionary.
        
        Args:
            data: Dictionary data
        """
        self.footnote_id = data.get('footnote_id', '')
        self.footnote_type = data.get('footnote_type', 'normal')
        self.content = data.get('content', '')
        self.properties = data.get('properties', {})
        self.validation_errors = data.get('validation_errors', [])
        
        logger.debug(f"Footnote loaded from dictionary: {self.footnote_id}")
    
    def get_footnote_info(self) -> Dict[str, Any]:
        """
        Get footnote information.
        
        Returns:
            Dictionary with footnote information
        """
        return {
            'footnote_id': self.footnote_id,
            'footnote_type': self.footnote_type,
            'content_length': len(self.content),
            'properties_count': len(self.properties),
            'is_valid': self.validate(),
            'validation_errors_count': len(self.validation_errors)
        }
    
    def clear_content(self) -> None:
        """Clear footnote content."""
        self.content = ""
        logger.debug("Footnote content cleared")
    
    def clear_properties(self) -> None:
        """Clear all footnote properties."""
        self.properties.clear()
        logger.debug("Footnote properties cleared")
    
    def has_property(self, key: str) -> bool:
        """
        Check if footnote has property.
        
        Args:
            key: Property key
            
        Returns:
            True if property exists, False otherwise
        """
        return key in self.properties
    
    def remove_property(self, key: str) -> bool:
        """
        Remove footnote property.
        
        Args:
            key: Property key to remove
            
        Returns:
            True if property was removed, False if not found
        """
        if key in self.properties:
            del self.properties[key]
            logger.debug(f"Footnote property removed: {key}")
            return True
        return False
    
    def is_empty(self) -> bool:
        """
        Check if footnote is empty.
        
        Returns:
            True if footnote has no content, False otherwise
        """
        return not self.content.strip()
    
    def get_length(self) -> int:
        """
        Get footnote content length.
        
        Returns:
            Length of footnote content
        """
        return len(self.content)


class Endnote(Models):
    """
    Represents an endnote in the document.
    
    Handles endnote functionality, properties, content, and validation.
    """
    
    def __init__(self, endnote_id: str = "", content: str = "", 
                 properties: Dict[str, Any] = None):
        """
        Initialize endnote.
        
        Args:
            endnote_id: Endnote identifier
            content: Endnote content
            properties: Endnote properties
        """
        super().__init__()
        self.endnote_id = endnote_id
        self.content = content
        self.properties = properties or {}
        self.validation_errors = []
        
        logger.debug(f"Endnote initialized: {endnote_id}")
    
    def set_endnote_id(self, endnote_id: str) -> None:
        """
        Set endnote ID.
        
        Args:
            endnote_id: Endnote identifier
        """
        if not endnote_id or not isinstance(endnote_id, str):
            raise ValueError("Endnote ID must be a non-empty string")
        
        self.endnote_id = endnote_id
        logger.debug(f"Endnote ID set to: {endnote_id}")
    
    def set_content(self, content: str) -> None:
        """
        Set endnote content.
        
        Args:
            content: Endnote content
        """
        if not isinstance(content, str):
            raise ValueError("Endnote content must be a string")
        
        self.content = content
        logger.debug(f"Endnote content set: {len(content)} characters")
    
    def get_endnote_id(self) -> str:
        """
        Get endnote ID.
        
        Returns:
            Endnote identifier
        """
        return self.endnote_id
    
    def get_content(self) -> str:
        """
        Get endnote content.
        
        Returns:
            Endnote content
        """
        return self.content
    
    def set_property(self, key: str, value: Any) -> None:
        """
        Set endnote property.
        
        Args:
            key: Property key
            value: Property value
        """
        self.properties[key] = value
        logger.debug(f"Endnote property set: {key} = {value}")
    
    def get_property(self, key: str, default: Any = None) -> Any:
        """
        Get endnote property.
        
        Args:
            key: Property key
            default: Default value if property not found
            
        Returns:
            Property value or default
        """
        return self.properties.get(key, default)
    
    def validate(self) -> bool:
        """
        Validate endnote.
        
        Returns:
            True if endnote is valid, False otherwise
        """
        self.validation_errors = []
        
        # Validate endnote ID
        if not self.endnote_id:
            self.validation_errors.append("Endnote ID is required")
        
        # Validate content
        if not self.content:
            self.validation_errors.append("Endnote content is required")
        
        # Validate properties
        if not isinstance(self.properties, dict):
            self.validation_errors.append("Endnote properties must be a dictionary")
        
        is_valid = len(self.validation_errors) == 0
        logger.debug(f"Endnote validation: {'valid' if is_valid else 'invalid'}")
        
        return is_valid
    
    def get_validation_errors(self) -> List[str]:
        """
        Get validation errors.
        
        Returns:
            List of validation errors
        """
        return self.validation_errors.copy()
    
    def to_dict(self) -> Dict[str, Any]:
        """
        Convert endnote to dictionary.
        
        Returns:
            Dictionary representation of endnote
        """
        return {
            'type': 'endnote',
            'endnote_id': self.endnote_id,
            'content': self.content,
            'properties': self.properties.copy(),
            'validation_errors': self.validation_errors.copy()
        }
    
    def from_dict(self, data: Dict[str, Any]) -> None:
        """
        Load endnote from dictionary.
        
        Args:
            data: Dictionary data
        """
        self.endnote_id = data.get('endnote_id', '')
        self.content = data.get('content', '')
        self.properties = data.get('properties', {})
        self.validation_errors = data.get('validation_errors', [])
        
        logger.debug(f"Endnote loaded from dictionary: {self.endnote_id}")
    
    def get_endnote_info(self) -> Dict[str, Any]:
        """
        Get endnote information.
        
        Returns:
            Dictionary with endnote information
        """
        return {
            'endnote_id': self.endnote_id,
            'content_length': len(self.content),
            'properties_count': len(self.properties),
            'is_valid': self.validate(),
            'validation_errors_count': len(self.validation_errors)
        }
    
    def clear_content(self) -> None:
        """Clear endnote content."""
        self.content = ""
        logger.debug("Endnote content cleared")
    
    def clear_properties(self) -> None:
        """Clear all endnote properties."""
        self.properties.clear()
        logger.debug("Endnote properties cleared")
    
    def has_property(self, key: str) -> bool:
        """
        Check if endnote has property.
        
        Args:
            key: Property key
            
        Returns:
            True if property exists, False otherwise
        """
        return key in self.properties
    
    def remove_property(self, key: str) -> bool:
        """
        Remove endnote property.
        
        Args:
            key: Property key to remove
            
        Returns:
            True if property was removed, False if not found
        """
        if key in self.properties:
            del self.properties[key]
            logger.debug(f"Endnote property removed: {key}")
            return True
        return False
    
    def is_empty(self) -> bool:
        """
        Check if endnote is empty.
        
        Returns:
            True if endnote has no content, False otherwise
        """
        return not self.content.strip()
    
    def get_length(self) -> int:
        """
        Get endnote content length.
        
        Returns:
            Length of endnote content
        """
        return len(self.content)



"""Utility functions for PDF generation."""

from typing import Tuple


def hex_to_rgb(hex_color: str, default: Tuple[float, float, float] = (0.0, 0.0, 0.0)) -> Tuple[float, float, float]:
    """Convert HEX color (#RRGGBB) to RGB tuple (0-1 scale).
    
    Args:
        hex_color: Color in HEX format (e.g., "#FF0000" or "FF0000")
        default: Default color to return if conversion fails
        
    Returns:
        Tuple of (r, g, b) values in 0-1 scale
    """
    # Handle special values
    if not hex_color or hex_color.lower() in ("auto", "none", "transparent"):
        return default
    
    # Remove # if present
    hex_color = hex_color.lstrip("#")
    
    # Validate hex color
    if len(hex_color) != 6:
        return default
    
    try:
        r = int(hex_color[0:2], 16) / 255.0
        g = int(hex_color[2:4], 16) / 255.0
        b = int(hex_color[4:6], 16) / 255.0
        return (r, g, b)
    except (ValueError, IndexError):
        return default


def escape_pdf_string(text: str) -> str:
    """Escape special characters in PDF strings.
    
    Args:
        text: Input string (will be converted to str if not already)
        
    Returns:
        Escaped string for PDF
    """
    # Handle None and non-string inputs
    if text is None:
        return ""
    if not isinstance(text, str):
        text = str(text)
    
    # PDF string escape: \n, \r, \t, \\, \(, \)
    replacements = {
        "\\": "\\\\",
        "(": "\\(",
        ")": "\\)",
        "\n": "\\n",
        "\r": "\\r",
        "\t": "\\t",
    }
    
    result = text
    for char, escaped in replacements.items():
        result = result.replace(char, escaped)
    
    return result


def format_pdf_number(value: float) -> str:
    """Format number for PDF (limit decimal places).
    
    Args:
        value: Numeric value
        
    Returns:
        Formatted string
    """
    # PDF typically uses 2-3 decimal places
    return f"{value:.3f}".rstrip("0").rstrip(".")


def format_pdf_matrix(a: float, b: float, c: float, d: float, e: float, f: float) -> str:
    """Format transformation matrix for PDF.
    
    Args:
        a, b, c, d, e, f: Matrix values
        
    Returns:
        Formatted matrix string
    """
    return f"{format_pdf_number(a)} {format_pdf_number(b)} {format_pdf_number(c)} {format_pdf_number(d)} {format_pdf_number(e)} {format_pdf_number(f)}"




"""Paragraph model for DOCX documents."""

from typing import List, Dict, Any, Optional, Tuple, Type, Union
from .base import Models
from .run import Run

class Paragraph(Models):
    """Represents a paragraph with runs, numbering, and formatting data."""
    
    def __init__(self):
        """Initialize paragraph."""
        super().__init__()
        self.runs: List[Run] = []
        self.style: Optional[Dict[str, Any]] = None
        self.numbering: Optional[Dict[str, Any]] = None
        self.tables: List[Any] = []  # inline tables
        # Paragraph formatting properties
        self.alignment: Optional[str] = None
        self.spacing_before: Optional[float] = None
        self.spacing_after: Optional[float] = None
        self.line_spacing: Optional[float] = None
        self.line_spacing_rule: Optional[str] = None
        self.spacing_before_lines: Optional[float] = None
        self.spacing_after_lines: Optional[float] = None
        self.spacing_before_auto: bool = False
        self.spacing_after_auto: bool = False
        self.left_indent: Optional[float] = None
        self.right_indent: Optional[float] = None
        self.first_line_indent: Optional[float] = None
        self.hanging_indent: Optional[float] = None
        self.borders: Optional[Dict[str, Any]] = None
        self.background: Optional[Dict[str, Any]] = None
        self.effect: Optional[Dict[str, Any]] = None
        self.outline: Optional[Dict[str, Any]] = None
        self.shadow: Optional[Dict[str, Any]] = None
        self.highlight: Optional[Dict[str, Any]] = None
        self.allowed_models: Tuple[Type[Models], ...] = (
            Run,  # Will be imported when needed
        )
    
    def add_run(self, run: Run):
        """Add run to paragraph."""
        # Accept both Run objects and dict representations
        if isinstance(run, dict):
            # Convert dict to Run object
            from .run import Run as RunClass
            run_obj = RunClass(
                text=run.get("text", ""),
                style=run.get("style"),
                space=run.get("space", "default"),
                has_break=run.get("has_break", False),
                has_tab=run.get("has_tab", False),
                has_drawing=run.get("has_drawing", False),
                break_type=run.get("break_type"),
            )
            # Copy footnote/endnote references if present
            if "footnote_refs" in run:
                run_obj.footnote_refs = run["footnote_refs"]
            if "endnote_refs" in run:
                run_obj.endnote_refs = run["endnote_refs"]
            self.runs.append(run_obj)
            self.add_child(run_obj)
        elif isinstance(run, Run):
            self.runs.append(run)
            self.add_child(run)
    
    def add_table(self, table):
        """Add inline table to paragraph."""
        self.tables.append(table)
        self.add_child(table)
    
    def add_field(self, field):
        """Add field to paragraph."""
        self.add_child(field)
    
    def add_hyperlink(self, hyperlink):
        """Add hyperlink to paragraph."""
        self.add_child(hyperlink)
    
    def set_style(self, style):
        """Set paragraph style."""
        self.style = style
    
    def set_numbering(self, numbering_ref):
        """Set paragraph numbering."""
        self.numbering = numbering_ref
    
    def set_list(self, level: int = 0, numbering_id: Optional[Union[str, int]] = None) -> None:
        """

        Sets paragraph as list element.

        Args:
        ...
        """
        if numbering_id is None:
            # If numbering_id not provided, use default
            # In future can add default numbering_id from document
            raise ValueError("numbering_id is required")
        
        # Convert numbering_id to string if needed
        if isinstance(numbering_id, int):
            numbering_id = str(numbering_id)
        elif hasattr(numbering_id, 'group_id'):
            # If this is NumberingGroup, use its group_id
            numbering_id = str(numbering_id.group_id)
        elif hasattr(numbering_id, 'num_id'):
            # If this has num_id (numbering instance)
            numbering_id = str(numbering_id.num_id)
        
        # Ustaw numbering jako dict zgodny z formatem DOCX
        self.numbering = {
            'id': numbering_id,
            'level': str(level)
        }
        
        # Also set in style if exists
        if self.style is None:
            self.style = {}
        
        if not isinstance(self.style, dict):
            self.style = {}
        
        self.style['numbering'] = {
            'id': numbering_id,
            'level': str(level)
        }
    
    def get_text(self, preserve_format=False):
        """Get text content from all runs."""
        text_parts = []
        for run in self.runs:
            run_text = run.get_text()
            if run_text:
                text_parts.append(run_text)
        return ' '.join(text_parts)
    
    def _merge_runs_plain(self):
        """Merge runs into plain text (like Word/py-docx)."""
        parts = []
        for run in self.runs:
            run_text = run.get_text()
            if run_text:
                parts.append(run_text)
        return " ".join(parts)
    
    def _merge_runs_formatted(self):
        """Merge runs into fragments that preserve formatting metadata."""
        fragments = []
        for run in self.runs:
            run_text = run.get_text()
            if run_text:
                fragments.append({
                    "text": run_text,
                    "style": getattr(run, 'style', {}),
                    "is_bold": getattr(run, 'is_bold', lambda: False)(),
                    "is_italic": getattr(run, 'is_italic', lambda: False)(),
                    "is_underline": getattr(run, 'is_underline', lambda: False)(),
                    "color": getattr(run, 'color', None)
                })
        return fragments
    
    def get_fragments(self):
        """Get formatted fragments for rich text export."""
        return self._merge_runs_formatted()
    
    def _normalize_runs(self):
        """Merge neighbouring runs when their style dictionaries match."""
        merged = []
        for run in self.runs:
            if merged and getattr(run, 'style', None) == getattr(merged[-1], 'style', None):
                merged[-1].text = (getattr(merged[-1], 'text', '') or '') + (getattr(run, 'text', '') or '')
            else:
                merged.append(run)
        self.runs = merged
    
    def create_inline_fragments(self):
        """Group runs into fragments with consistent style metadata."""
        fragments = []
        current_fragment = None
        
        for run in self.runs:
            if current_fragment is None or not self._runs_have_same_style(current_fragment, run):
                current_fragment = {
                    'runs': [],
                    'style': getattr(run, 'style', {})
                }
                fragments.append(current_fragment)
            
            current_fragment['runs'].append(run)
        
        return fragments
    
    def _runs_have_same_style(self, run1, run2):
        """Check if two runs have the same style."""
        style1 = getattr(run1, 'style', {})
        style2 = getattr(run2, 'style', {})
        return style1 == style2
    
    def get_runs(self):
        """Get all runs in paragraph."""
        return self.runs.copy()
    
    def is_list_item(self):
        """Check if paragraph is a list item."""
        return self.numbering is not None
    
    @property
    def text(self):
        """Return merged plain text similar to Word/py-docx interfaces."""
        return self._merge_runs_plain()
    
    def set_text(self, text: str):
        """
        Set text content for paragraph.
        
        Args:
            text: Text to set
        """
        # Clear existing runs
        self.runs.clear()
        self.children.clear()
        
        # Create a new run with the text
        run = Run()
        run.text = text
        self.add_run(run)


"""
JSON exporter for DOCX documents.

Handles JSON export with full model serialization and formatting.
"""

import json
from typing import Dict, Any, Optional, List, Union
from pathlib import Path
import logging
from datetime import datetime

logger = logging.getLogger(__name__)

class JSONExporter:
    """
    Exports document as JSON (full model).
    
    Handles JSON export with full model serialization and formatting.
    """
    
    def __init__(self, document: Any, indent: int = 2, ensure_ascii: bool = False):
        """
        Initialize JSON exporter.
        
        Args:
            document: Document to export
            indent: JSON indentation level
            ensure_ascii: Whether to ensure ASCII encoding
        """
        self.document = document
        self.indent = indent
        self.ensure_ascii = ensure_ascii
        
        logger.debug("JSON exporter initialized")
    
    def export(self, output_path: Union[str, Path]) -> bool:
        """
        Export document to JSON.
        
        Args:
            output_path: Output file path
            
        Returns:
            True if export successful, False otherwise
        """
        try:
            output_path = Path(output_path)
            output_path.parent.mkdir(parents=True, exist_ok=True)
            
            # Export document model
            document_data = self.export_document_model(self.document)
            
            # Format JSON output
            json_output = self.format_json_output(document_data)
            
            # Write to file
            with open(output_path, 'w', encoding='utf-8') as f:
                f.write(json_output)
            
            logger.info(f"Document exported to JSON: {output_path}")
            return True
            
        except Exception as e:
            logger.error(f"Failed to export document to JSON: {e}")
            return False
    
    def export_document_model(self, document: Any = None) -> Dict[str, Any]:
        """
        Export full document model.
        
        Args:
            document: Document to export (uses self.document if not provided)
            
        Returns:
            Document model as dictionary
        """
        if document is None:
            document = self.document
        
        document_data = {
            'metadata': {
                'export_timestamp': datetime.now().isoformat(),
                'exporter_version': '1.0.0',
                'document_type': 'DOCX'
            },
            'content': self._extract_document_content(document),
            'styles': self._extract_document_styles(document),
            'layout': self._extract_document_layout(document),
            'document': {
                'properties': self._extract_document_properties(document),
                'content': self._extract_document_content(document),
                'styles': self._extract_document_styles(document),
                'layout': self._extract_document_layout(document)
            }
        }
        
        logger.debug("Document model exported")
        return document_data
    
    def export_element(self, element: Any) -> Dict[str, Any]:
        """
        Export individual element.
        
        Args:
            element: Element to export
            
        Returns:
            Element as dictionary
        """
        if hasattr(element, 'to_dict'):
            return element.to_dict()
        elif hasattr(element, '__dict__'):
            return self._serialize_object(element)
        else:
            return {'type': type(element).__name__, 'value': str(element)}
    
    def format_json_output(self, data: Dict[str, Any]) -> str:
        """
        Format JSON output.
        
        Args:
            data: Data to format
            
        Returns:
            Formatted JSON string
        """
        try:
            json_output = json.dumps(
                data,
                indent=self.indent,
                ensure_ascii=self.ensure_ascii,
                default=self._json_serializer
            )
            return json_output
        except Exception as e:
            logger.error(f"Failed to format JSON: {e}")
            return json.dumps({'error': str(e)})
    
    def _extract_document_properties(self, document: Any) -> Dict[str, Any]:
        """Extract document properties."""
        properties = {}
        
        if hasattr(document, 'metadata'):
            properties['metadata'] = self._serialize_object(document.metadata)
        
        if hasattr(document, 'title'):
            properties['title'] = document.title
        
        if hasattr(document, 'author'):
            properties['author'] = document.author
        
        if hasattr(document, 'created'):
            properties['created'] = document.created
        
        if hasattr(document, 'modified'):
            properties['modified'] = document.modified
        
        return properties
    
    def _extract_document_content(self, document: Any) -> Dict[str, Any]:
        """Extract document content."""
        content = {
            'paragraphs': [],
            'tables': [],
            'images': [],
            'other_elements': []
        }
        
        # Extract paragraphs
        if hasattr(document, 'get_paragraphs'):
            paragraphs = document.get_paragraphs()
            for para in paragraphs:
                content['paragraphs'].append(self.export_element(para))
        
        # Extract tables
        if hasattr(document, 'get_tables'):
            tables = document.get_tables()
            for table in tables:
                content['tables'].append(self.export_element(table))
        
        # Extract images
        if hasattr(document, 'get_images'):
            images = document.get_images()
            for image in images:
                content['images'].append(self.export_element(image))
        
        return content
    
    def _extract_document_styles(self, document: Any) -> Dict[str, Any]:
        """Extract document styles."""
        styles = {}
        
        if hasattr(document, 'styles'):
            styles['document_styles'] = self._serialize_object(document.styles)
        
        if hasattr(document, 'themes'):
            styles['themes'] = self._serialize_object(document.themes)
        
        return styles
    
    def _extract_document_layout(self, document: Any) -> Dict[str, Any]:
        """Extract document layout."""
        layout = {}
        
        if hasattr(document, 'pages'):
            layout['pages'] = []
            for page in document.pages:
                layout['pages'].append(self.export_element(page))
        
        if hasattr(document, 'sections'):
            layout['sections'] = []
            for section in document.sections:
                layout['sections'].append(self.export_element(section))
        
        return layout
    
    def _serialize_object(self, obj: Any) -> Dict[str, Any]:
        """Serialize object to dictionary."""
        if obj is None:
            return None
        
        if isinstance(obj, (str, int, float, bool)):
            return obj
        
        if isinstance(obj, (list, tuple)):
            return [self._serialize_object(item) for item in obj]
        
        if isinstance(obj, dict):
            return {key: self._serialize_object(value) for key, value in obj.items()}
        
        if hasattr(obj, '__dict__'):
            result = {}
            for key, value in obj.__dict__.items():
                if not key.startswith('_'):
                    result[key] = self._serialize_object(value)
            return result
        
        return str(obj)
    
    def _json_serializer(self, obj: Any) -> Any:
        """JSON serializer for non-standard types."""
        if hasattr(obj, 'isoformat'):  # datetime objects
            return obj.isoformat()
        elif hasattr(obj, '__dict__'):  # custom objects
            return self._serialize_object(obj)
        else:
            return str(obj)
    
    def export_to_string(self) -> str:
        """
        Export document to JSON string.
        
        Returns:
            JSON string representation
        """
        document_data = self.export_document_model(self.document)
        return self.format_json_output(document_data)
    
    def validate_json(self, json_string: str) -> bool:
        """
        Validate JSON string.
        
        Args:
            json_string: JSON string to validate
            
        Returns:
            True if valid JSON, False otherwise
        """
        try:
            json.loads(json_string)
            return True
        except json.JSONDecodeError:
            return False
    
    def get_export_info(self) -> Dict[str, Any]:
        """Get export information."""
        return {
            'exporter_type': 'JSON',
            'indent': self.indent,
            'ensure_ascii': self.ensure_ascii,
            'document_type': type(self.document).__name__ if self.document else None
        }



"""
Rust-based PDF compiler using pdf-writer.

This module provides a high-performance PDF renderer using Rust (pdf-writer)
as an alternative to ReportLab. It can be used as a drop-in replacement
or as a fallback option.
"""

import logging
import mimetypes
import threading
import tempfile
import hashlib
import os
import uuid
from pathlib import Path
from typing import Optional, Dict, Any, List, Set, Tuple
from dataclasses import asdict, fields, is_dataclass

logger = logging.getLogger(__name__)
_IMAGE_TARGET_DPI = 192.0  # Match ReportLab implementation for rasterizing EMF/WMF

# Try to import Rust renderer
pdf_renderer_rust = None
HAS_RUST_RENDERER = False

# Try docquill_rust first (new unified package from PyPI)
try:
    import docquill_rust
    if hasattr(docquill_rust, 'PdfRenderer'):
        pdf_renderer_rust = docquill_rust
        HAS_RUST_RENDERER = True
        logger.debug("Loaded Rust renderer from docquill_rust (PdfRenderer)")
except ImportError:
    pass

# Legacy: try pdf_renderer_rust
if not HAS_RUST_RENDERER:
    try:
        import pdf_renderer_rust as _pdf_renderer_rust
        if hasattr(_pdf_renderer_rust, 'PdfRenderer'):
            pdf_renderer_rust = _pdf_renderer_rust
            HAS_RUST_RENDERER = True
            logger.debug("Loaded Rust renderer from pdf_renderer_rust")
    except ImportError:
        pass

if not HAS_RUST_RENDERER:
    logger.debug(
        "pdf_renderer_rust not available. Install docquill-rust: pip install docquill-rust"
    )

try:
    from ..unified_layout import UnifiedLayout, LayoutPage, LayoutBlock
    from ..geometry import Rect, Size, Margins, emu_to_points
    from ..layout_primitives import (
        BlockContent,
        ParagraphLayout,
        ParagraphLine,
        InlineBox,
        OverlayBox,
    )
except ImportError:
    # Allow running as a script for testing (add parent to path)
    import sys
    from pathlib import Path
    project_root = Path(__file__).parent.parent.parent.parent
    sys.path.insert(0, str(project_root))
    from docquill.engine.unified_layout import UnifiedLayout, LayoutPage, LayoutBlock
    from docquill.engine.geometry import Rect, Size, Margins, emu_to_points
    from docquill.engine.layout_primitives import (
        BlockContent,
        ParagraphLayout,
        ParagraphLine,
        InlineBox,
        OverlayBox,
    )
from ...media import MediaConverter


class PDFCompilerRust:
    """
    Rust-based PDF compiler using pdf-writer.
    
    Provides high-performance PDF rendering as an alternative to ReportLab.
    """
    
    def __init__(
        self,
        output_path: str,
        page_size: tuple = (595.0, 842.0),  # A4 in points
        package_reader: Optional[Any] = None,
        footnote_renderer: Optional[Any] = None,
        image_cache: Optional[Any] = None,
        media_converter: Optional[Any] = None,
        watermark_opacity: Optional[float] = None,
    ):
        """
        Initialize Rust PDF compiler.
        
        Args:
            output_path: Path to output PDF file
            page_size: Page size tuple (width, height) in points
            package_reader: PackageReader instance (for image paths)
            footnote_renderer: FootnoteRenderer instance (optional)
            image_cache: LayoutPipeline image cache (for pre-converted WMF/EMF images)
            media_converter: MediaConverter instance (optional)
            watermark_opacity: Optional opacity override for watermark blocks (0.0-1.0)
        """
        if not HAS_RUST_RENDERER:
            raise ImportError(
                "Rust PDF renderer not available. "
                "Install with: pip install docquill-rust"
            )
        
        self.output_path = str(output_path)
        self.page_size = page_size
        self.package_reader = package_reader
        self.footnote_renderer = footnote_renderer
        self.image_cache = image_cache
        self.media_converter = media_converter or MediaConverter()
        self._temp_files: List[Path] = []
        self._converted_images: Dict[str, Path] = {}
        self._current_timings: Optional[Dict[str, List[float]]] = None
        self._registered_stream_keys: Set[str] = set()
        self.image_streams: Dict[str, bytes] = {}  # Store image bytes for streaming
        self.watermark_opacity_override = self._normalize_opacity_value(watermark_opacity)
        
        # Initialize Rust renderer
        self.renderer = pdf_renderer_rust.PdfRenderer(
            self.output_path,
            page_size[0],
            page_size[1],
        )
        
        # Performance limits (can be adjusted)
        self._max_pages: Optional[int] = None  # None = no limit
        self._max_blocks_per_page: Optional[int] = None  # None = no limit
        self._max_lines_per_paragraph: Optional[int] = None  # None = no limit
        
        # Enable compact layout by default
        self._use_compact_layout = True

        # Dispatcher: block_type -> handler
        self._dispatch = self._build_dispatcher()
        
        logger.info(f"PDFCompilerRust initialized: output={self.output_path}, page_size={self.page_size}")
    
    def __del__(self):
        for tmp in getattr(self, "_temp_files", []) or []:
            try:
                Path(tmp).unlink(missing_ok=True)
            except Exception:
                pass

    # ======================================================================
    # Dispatcher
    # ======================================================================

    def _build_dispatcher(self) -> Dict[str, Any]:
        """
        Build dispatcher mapping block_type -> handler function.
        Each handler has signature:
            handler(x, y, width, height, content, block)
        """
        return {
            "paragraph": self._handle_paragraph_block,
            "table": self._handle_table_block,
            "image": self._handle_image_block,
            "header": self._handle_header_block,
            "footer": self._handle_footer_block,
            "footnotes": self._handle_footnotes_block,
            "endnotes": self._handle_endnotes_block,
            # Shapes / textboxes / decorations
            "textbox": self._handle_rect_block,
            "decorator": self._handle_rect_block,
            "rectangle": self._handle_rect_block,
            "rect": self._handle_rect_block,
            "vml_shape": self._handle_rect_block,
        }

    def _render_element(self, block: LayoutBlock) -> None:
        """
        Central dispatch for a single block using dict lookup instead of
        match/case spaghetti.
        """
        block_type = getattr(block, "block_type", None)
        handler = self._dispatch.get(block_type)

        if handler is None:
            logger.warning(f"Unknown block type: {block_type}, skipping")
            return

        frame = block.frame
        x, y, width, height = frame.x, frame.y, frame.width, frame.height
        
        content = self._prepare_content_for_renderer(block.content)
        
        # For paragraphs, check if layout_payload is in block or content
        if block_type == "paragraph":
            # First check if block has layout_payload directly
            if hasattr(block, "layout_payload") and block.layout_payload:
                layout_dict = self._safe_to_dict(block.layout_payload, visited=None)
                if "lines" in layout_dict:
                    content["layout_payload"] = layout_dict
                    logger.info(f"✅ Extracted layout_payload from block.layout_payload for paragraph at ({x:.1f}, {y:.1f}), lines: {len(layout_dict.get('lines', []))}")
            elif hasattr(block, "_layout_payload") and block._layout_payload:
                layout_dict = self._safe_to_dict(block._layout_payload, visited=None)
                if "lines" in layout_dict:
                    content["layout_payload"] = layout_dict
                    logger.info(f"✅ Extracted layout_payload from block._layout_payload for paragraph at ({x:.1f}, {y:.1f}), lines: {len(layout_dict.get('lines', []))}")
            
            # Then check if content has layout_payload
            if "layout_payload" not in content and "_layout_payload" not in content:
                # Check if content has payload with layout_payload
                if "payload" in content:
                    payload = content["payload"]
                    logger.warning(f"🔍 Paragraph at ({x:.1f}, {y:.1f}): payload type: {type(payload).__name__}")
                    # payload might be dict or ParagraphLayout object
                    if isinstance(payload, dict):
                        logger.warning(f"  payload keys: {list(payload.keys())[:10]}")
                        # Check if payload has lines directly or layout_payload
                        if "lines" in payload:
                            # payload is ParagraphLayout with lines
                            content["layout_payload"] = payload
                            logger.info(f"✅ Extracted layout_payload from payload (dict) for paragraph at ({x:.1f}, {y:.1f}), lines: {len(payload.get('lines', []))}")
                        elif "data" in payload:
                            # Check if payload.data has lines
                            data = payload["data"]
                            logger.warning(f"  payload.data type: {type(data).__name__}")
                            if isinstance(data, dict):
                                logger.warning(f"  payload.data keys: {list(data.keys())[:10]}")
                                if "lines" in data:
                                    content["layout_payload"] = data
                                    logger.info(f"✅ Extracted layout_payload from payload.data for paragraph at ({x:.1f}, {y:.1f}), lines: {len(data.get('lines', []))}")
                            elif hasattr(data, "lines"):
                                # data is ParagraphLayout object
                                data_dict = self._safe_to_dict(data, visited=None)
                                if "lines" in data_dict:
                                    content["layout_payload"] = data_dict
                                    logger.info(f"✅ Extracted layout_payload from payload.data (object) for paragraph at ({x:.1f}, {y:.1f}), lines: {len(data_dict.get('lines', []))}")
                        elif "layout_payload" in payload:
                            content["layout_payload"] = payload["layout_payload"]
                            logger.info(f"✅ Extracted layout_payload from payload.layout_payload for paragraph at ({x:.1f}, {y:.1f})")
                    elif hasattr(payload, "lines"):
                        # payload is ParagraphLayout object - convert to dict
                        logger.warning(f"  payload is object with lines attribute, converting...")
                        payload_dict = self._safe_to_dict(payload, visited=None)
                        logger.warning(f"  converted payload keys: {list(payload_dict.keys())[:10]}")
                        if "lines" in payload_dict:
                            content["layout_payload"] = payload_dict
                            logger.info(f"✅ Extracted layout_payload from payload (object) for paragraph at ({x:.1f}, {y:.1f}), lines: {len(payload_dict.get('lines', []))}")
                        else:
                            logger.warning(f"⚠️ payload object has no lines after conversion, keys: {list(payload_dict.keys())}")
                    else:
                        logger.warning(f"⚠️ payload is neither dict nor has lines attribute, type: {type(payload).__name__}")
                else:
                    logger.warning(f"⚠️ Paragraph at ({x:.1f}, {y:.1f}): no 'payload' in content")
            
            # Final check
            has_layout = bool(content.get("layout_payload") or content.get("_layout_payload"))
            if has_layout:
                layout = content.get("layout_payload") or content.get("_layout_payload")
                has_lines = bool(layout and isinstance(layout, dict) and layout.get("lines"))
                if not has_lines:
                    logger.warning(f"⚠️ Paragraph at ({x:.1f}, {y:.1f}) has layout_payload but no lines! layout keys: {list(layout.keys()) if isinstance(layout, dict) else 'not dict'}")
            else:
                logger.warning(f"⚠️ Paragraph at ({x:.1f}, {y:.1f}) has no layout_payload, content keys: {list(content.keys())[:10]}")

        is_watermark = bool(content.get("is_watermark"))

        if is_watermark:
            opacity = self._get_watermark_opacity(block, content)
            self.renderer.canvas_save_state()
            try:
                if opacity < 0.999:
                    self.renderer.canvas_set_opacity(opacity)
                handler(x, y, width, height, content, block)
            finally:
                self.renderer.canvas_restore_state()
        else:
            handler(x, y, width, height, content, block)

    def _get_watermark_opacity(self, block: LayoutBlock, content: Dict[str, Any]) -> float:
        """
        Determine the opacity for a watermark block.
        """
        if self.watermark_opacity_override is not None:
            return self.watermark_opacity_override

        explicit = self._extract_opacity_value(content, include_generic=True)
        if explicit is None and hasattr(block, "style") and isinstance(block.style, dict):
            explicit = self._extract_opacity_value(block.style, include_generic=False)
        if explicit is not None:
            return max(0.0, min(1.0, explicit))

        block_type = getattr(block, "block_type", "") or ""
        content_type = content.get("type") or content.get("content", {}).get("type", "")
        normalized_type = str(content_type).lower()
        if block_type == "image" or normalized_type == "image":
            return 0.5
        if normalized_type == "vml_shape":
            return 0.3
        return 0.35

    def _extract_opacity_value(self, value: Any, include_generic: bool = True) -> Optional[float]:
        """
        Recursively search for an opacity hint inside a nested content dict/list.
        """
        try_keys = ("watermark_opacity", "opacity") if include_generic else ("watermark_opacity",)
        if isinstance(value, dict):
            for key in try_keys:
                if key in value:
                    try:
                        normalized = float(value[key])
                        return max(0.0, min(1.0, normalized))
                    except (TypeError, ValueError):
                        continue
            for child in value.values():
                result = self._extract_opacity_value(child)
                if result is not None:
                    return result
        elif isinstance(value, (list, tuple)):
            for item in value:
                result = self._extract_opacity_value(item)
                if result is not None:
                    return result
        return None

    @staticmethod
    def _normalize_opacity_value(value: Optional[float]) -> Optional[float]:
        if value is None:
            return None
        try:
            normalized = float(value)
        except (TypeError, ValueError):
            logger.warning(f"Ignoring invalid watermark opacity value: {value}")
            return None
        return max(0.0, min(1.0, normalized))

    # ======================================================================
    # Public entry
    # ======================================================================
    
    def compile(self, unified_layout: UnifiedLayout) -> str:
        """
        Compile UnifiedLayout to PDF using dispatcher-based iteration.
        Iterates over all elements and calls Rust methods with x, y, width, height.
        
        Args:
            unified_layout: UnifiedLayout instance to render
            
        Returns:
            Path to generated PDF file
        """
        import time as time_module
        start_time = time_module.perf_counter()
        timings = {
            "content_prep": 0.0,
            "rust_calls": 0.0,
            "save": 0.0,
            "total": 0.0,
        }
        
        try:
            self._registered_stream_keys.clear()
            # Set total pages for field codes
            total_pages = max(len(unified_layout.pages), 1)
            self.renderer.set_total_pages(total_pages)
            
            logger.info(f"🔄 Rendering {len(unified_layout.pages)} pages (dispatcher-based)...")
            
            # Iterate through pages and blocks, calling Rust methods via dispatcher
            for page in unified_layout.pages:
                # Set current page number
                self.renderer.set_current_page_number(page.number)
                
                # Create new page
                self.renderer.new_page(page.size.width, page.size.height)
                
                # Sort blocks: watermarks first, then headers, body, footnotes, footers
                watermark_blocks = []
                header_blocks = []
                body_blocks = []
                footnote_blocks = []
                footer_blocks = []
                
                t0 = time_module.perf_counter()
                for block in page.blocks:
                    # Get content as dict-ish
                    content = block.content
                    if not isinstance(content, dict):
                        if hasattr(content, "__dict__"):
                            content = {k: v for k, v in content.__dict__.items() if not k.startswith("__")}
                        elif hasattr(content, "to_dict"):
                            content = content.to_dict()
                        else:
                            content = {}
                    
                    header_footer_context = content.get("header_footer_context")
                    
                    # Classification
                    if content.get("is_watermark") or (block.block_type == "textbox" and header_footer_context == "header"):
                        watermark_blocks.append(block)
                    elif block.block_type == "header" or header_footer_context == "header":
                        header_blocks.append(block)
                    elif block.block_type == "footer" or header_footer_context == "footer":
                        footer_blocks.append(block)
                    elif block.block_type == "footnotes":
                        footnote_blocks.append(block)
                    else:
                        body_blocks.append(block)
                timings["content_prep"] += time_module.perf_counter() - t0
                
                # Render in order: watermarks, headers, body, footnotes, footers
                t0 = time_module.perf_counter()
                for block in watermark_blocks:
                    self._render_element(block)
                
                for block in header_blocks:
                    self._render_element(block)
                
                for block in body_blocks:
                    self._render_element(block)
                
                for block in footnote_blocks:
                    self._render_element(block)
                
                for block in footer_blocks:
                    self._render_element(block)
                timings["rust_calls"] += time_module.perf_counter() - t0
            
            elapsed = time_module.perf_counter() - start_time
            logger.info(f"✅ Layout rendered in {elapsed:.2f}s")
            
            # Save PDF
            logger.info(f"💾 Saving PDF to: {self.output_path}")
            output_path = Path(self.output_path)
            output_path.parent.mkdir(parents=True, exist_ok=True)
            
            t0 = time_module.perf_counter()
            self.renderer.save()
            timings["save"] = time_module.perf_counter() - t0
            logger.info("✅ PDF save() completed")
            
            timings["total"] = time_module.perf_counter() - start_time
            logger.info(f"⏱️  Rust compile timings: content_prep={timings['content_prep']:.3f}s, rust_calls={timings['rust_calls']:.3f}s, save={timings['save']:.3f}s, total={timings['total']:.3f}s")
            import sys
            sys.stdout.write(f"RUST_TIMINGS:content_prep={timings['content_prep']:.6f},rust_calls={timings['rust_calls']:.6f},save={timings['save']:.6f},total={timings['total']:.6f}\n")
            sys.stdout.flush()
            
            # Verify file was created
            if output_path.exists():
                file_size = output_path.stat().st_size
                if file_size > 0:
                    logger.info("Registered %d unique image streams", len(self._registered_stream_keys))
                    logger.info(
                        f"✅ PDF generated successfully: {output_path.absolute()} "
                        f"({file_size:,} bytes)"
                    )
                    return str(output_path.absolute())
                else:
                    error_msg = f"PDF file was created but is empty (0 bytes): {output_path.absolute()}"
                    logger.error(error_msg)
                    raise RuntimeError(error_msg)
            else:
                error_msg = f"PDF file was not created: {output_path.absolute()}"
                logger.error(error_msg)
                raise RuntimeError(error_msg)
                
        except Exception as e:
            logger.error(f"❌ Rust PDF compilation failed: {e}", exc_info=True)
            raise
    
    # ======================================================================
    # Handlers used by dispatcher
    # ======================================================================

    def _handle_paragraph_block(
        self,
        x: float,
        y: float,
        width: float,
        height: float,
        content: Dict[str, Any],
        block: LayoutBlock,
    ) -> None:
        logger.debug(f"📄 Rendering paragraph at ({x:.1f}, {y:.1f}), size {width:.1f}x{height:.1f}")
        has_layout = bool(content.get("layout_payload") or content.get("_layout_payload"))
        logger.debug(f"  Has layout_payload: {has_layout}")
        self._render_paragraph_with_lines(x, y, width, height, content)

    def _handle_table_block(
        self,
        x: float,
        y: float,
        width: float,
        height: float,
        content: Dict[str, Any],
        block: LayoutBlock,
    ) -> None:
        rows = content.get("rows", [])
        logger.info(f"📊 Rendering table at ({x:.1f}, {y:.1f}), size {width:.1f}x{height:.1f}, rows: {len(rows)}")
        self._render_table_with_cells(x, y, width, height, content)

    def _handle_image_block(
        self,
        x: float,
        y: float,
        width: float,
        height: float,
        content: Dict[str, Any],
        block: LayoutBlock,
    ) -> None:
        self._render_image_direct(x, y, width, height, content)

    def _handle_header_block(
        self,
        x: float,
        y: float,
        width: float,
        height: float,
        content: Dict[str, Any],
        block: LayoutBlock,
    ) -> None:
        # Headers can be paragraph or more complex block
        if isinstance(content, dict):
            layout_payload = content.get("layout_payload") or content.get("_layout_payload")
            if layout_payload and isinstance(layout_payload, dict) and layout_payload.get("lines"):
                self._render_paragraph_with_lines(x, y, width, height, content)
                return
        self.renderer.render_header_block(x, y, width, height, content)

    def _handle_footer_block(
        self,
        x: float,
        y: float,
        width: float,
        height: float,
        content: Dict[str, Any],
        block: LayoutBlock,
    ) -> None:
        if isinstance(content, dict):
            layout_payload = content.get("layout_payload") or content.get("_layout_payload")
            if layout_payload and isinstance(layout_payload, dict) and layout_payload.get("lines"):
                self._render_paragraph_with_lines(x, y, width, height, content)
                return
        self.renderer.render_footer_block(x, y, width, height, content)

    def _handle_footnotes_block(
        self,
        x: float,
        y: float,
        width: float,
        height: float,
        content: Dict[str, Any],
        block: LayoutBlock,
    ) -> None:
        # Extract footnotes list from content - no JSON, just values
        footnotes_list = []
        if isinstance(content, dict):
            footnotes = content.get("footnotes", [])
            if isinstance(footnotes, list):
                for footnote in footnotes:
                    if isinstance(footnote, dict):
                        number = str(footnote.get("number", "?"))
                        footnote_content = str(footnote.get("content", ""))
                        footnotes_list.append((number, footnote_content))
        
        logger.info(f"📝 Rendering {len(footnotes_list)} footnotes at ({x:.1f}, {y:.1f}), size {width:.1f}x{height:.1f}")
        if footnotes_list:
            for i, (num, text) in enumerate(footnotes_list[:3]):  # Log first 3
                logger.info(f"  Footnote {i+1}: {num} = {text[:50]}...")
        
        # Call Rust with direct values, no JSON
        self.renderer.render_footnotes_block(x, y, width, height, footnotes_list)

    def _handle_endnotes_block(
        self,
        x: float,
        y: float,
        width: float,
        height: float,
        content: Dict[str, Any],
        block: LayoutBlock,
    ) -> None:
        self.renderer.render_endnotes_block(x, y, width, height, content)

    def _handle_rect_block(
        self,
        x: float,
        y: float,
        width: float,
        height: float,
        content: Dict[str, Any],
        block: LayoutBlock,
    ) -> None:
        """
        Generic rectangle/textbox/decorator/vml_shape handler.
        """
        style_dict = self._style_to_dict(getattr(block, "style", None))
        fill_color = (
            style_dict.get("fill_color")
            or style_dict.get("background_color")
        )
        stroke_color = (
            style_dict.get("stroke_color")
            or style_dict.get("border_color")
            or style_dict.get("color")
        )
        line_width = (
            style_dict.get("line_width")
            or style_dict.get("border_width")
            or 1.0
        )
        self.renderer.render_rectangle_block(
            x,
            y,
            width,
            height,
            str(fill_color) if fill_color else None,
            str(stroke_color) if stroke_color else None,
            line_width,
        )

    # ======================================================================
    # Paragraph rendering
    # ======================================================================

    def _render_paragraph_with_lines(
        self,
        x: float,
        y: float,
        width: float,
        height: float,
        content: Dict[str, Any],
    ) -> None:
        """
        Render paragraph by iterating over lines and items, calling Rust draw_string
        or image methods for each item (once exposed in Rust).
        """
        # Get layout_payload with lines
        layout_payload = content.get("layout_payload") or content.get("_layout_payload")
        if not layout_payload or not isinstance(layout_payload, dict):
            logger.warning(f"⚠️ Paragraph at ({x:.1f}, {y:.1f}) has no layout_payload with lines, skipping")
            logger.debug(f"  Content keys: {list(content.keys())[:10]}")
            return
        
        lines = layout_payload.get("lines", [])
        if not lines:
            logger.warning(f"⚠️ Paragraph at ({x:.1f}, {y:.1f}) has no lines in layout_payload, skipping")
            logger.debug(f"  layout_payload keys: {list(layout_payload.keys())}")
            return
        
        logger.info(f"📝 Rendering paragraph at ({x:.1f}, {y:.1f}) with {len(lines)} lines")
        
        # Call Rust render_paragraph_block which handles layout_payload
        # This method expects content with layout_payload.lines
        self.renderer.render_paragraph_block(x, y, width, height, content)
        return  # Skip the manual iteration below - Rust handles it
        
        # Get paragraph style (for future use if needed)
        style = content.get("style", {}) or {}
        padding = style.get("padding", {})
        if isinstance(padding, (list, tuple)) and len(padding) >= 4:
            padding_top, padding_right, padding_bottom, padding_left = (
                padding[0],
                padding[1],
                padding[2],
                padding[3],
            )
        else:
            padding_top = padding.get("top", 0.0) if isinstance(padding, dict) else 0.0
            padding_right = padding.get("right", 0.0) if isinstance(padding, dict) else 0.0
            padding_bottom = padding.get("bottom", 0.0) if isinstance(padding, dict) else 0.0
            padding_left = padding.get("left", 0.0) if isinstance(padding, dict) else 0.0
        
        text_left = x + padding_left
        text_top = y + height - padding_top  # PDF: Y increases upward
        
        # Iterate over lines and items
        for line in lines:
            if not isinstance(line, dict):
                    continue
            
            baseline_y = line.get("baseline_y", 0.0)
            offset_x = line.get("offset_x", 0.0)
            line_y = text_top - baseline_y  # PDF: Y increases upward
            
            items = line.get("items", [])
            for item in items:
                if not isinstance(item, dict):
                        continue
                
                item_kind = item.get("kind", "text_run")
                item_x = item.get("x", 0.0)
                item_data = item.get("data", {})
                
                match item_kind:
                    case "text_run" | "field":
                        # Extract text
                        text = item_data.get("text") if isinstance(item_data, dict) else None
                        if not text:
                            text = item.get("text")
                        if not text:
                            continue
                        
                        # Get item style
                        item_style = item_data.get("style") if isinstance(item_data, dict) else {}
                        font_name = item_style.get("font_name") or style.get("font_name", "DejaVu Sans")
                        font_size = item_style.get("font_size") or style.get("font_size", 11.0)
                        color = item_style.get("color") or style.get("color", "#000000")
                        
                        # Calculate position
                        text_x = text_left + offset_x + item_x
                        text_y = line_y
                        
                        # Use Rust render_paragraph_block which handles layout_payload
                        # For now, we'll call it once per paragraph with full layout_payload
                        # TODO: Add direct draw_string method to Rust for fine-grained control
                        pass  # Will be handled by render_paragraph_block call below
                    
                    case "inline_image" | "image":
                        # Extract image data
                        image_data = item_data.get("image") if isinstance(item_data, dict) else item_data
                        if not isinstance(image_data, dict):
                            continue
                        
                        stream_key = image_data.get("stream_key")
                        image_path = image_data.get("path") or image_data.get("image_path")
                        item_width = item.get("width", 0.0)
                        item_height = item.get("height", 0.0)
                        item_ascent = item.get("ascent", 0.0)
                        item_descent = item.get("descent", 0.0)
                        
                        # Calculate image position and size
                        image_x = text_left + offset_x + item_x
                        image_y = line_y - (item_ascent + item_descent)  # PDF: Y increases upward
                        image_width = item_width if item_width > 0 else 50.0
                        image_height = (
                            (item_ascent + item_descent)
                            if (item_ascent > 0 or item_descent > 0)
                            else item_height if item_height > 0 else 50.0
                        )
                        
                        # TODO: Replace with Rust draw_image method when available
                        logger.debug(
                            f"Would draw image at ({image_x:.2f}, {image_y:.2f}) "
                            f"size ({image_width:.2f}, {image_height:.2f})"
                        )
                    
                    case _:
                        logger.debug(f"Unknown item kind: {item_kind}, skipping")
        
        # Render overlays if any
        overlays = layout_payload.get("overlays", [])
        for overlay in overlays:
            if not isinstance(overlay, dict):
                continue
            self._render_overlay(overlay)

    # ======================================================================
    # Table rendering
    # ======================================================================

    def _render_table_with_cells(
        self,
        x: float,
        y: float,
        width: float,
        height: float,
        content: Dict[str, Any],
    ) -> None:
        """
        Render table by iterating over rows -> cells -> content, calling Rust
        methods for each element.
        """
        rows = content.get("rows", [])
        if not rows:
            logger.warning("Table has no rows, skipping")
            return
        
        # Iterate over rows and cells
        for row in rows:
            if not isinstance(row, dict):
                continue
            
            cells = row.get("cells", [])
            for cell in cells:
                if not isinstance(cell, dict):
                    continue
                
                cell_frame = cell.get("frame", {})
                if isinstance(cell_frame, dict):
                    cell_x = cell_frame.get("x", x)
                    cell_y = cell_frame.get("y", y)
                    cell_width = cell_frame.get("width", width)
                    cell_height = cell_frame.get("height", height)
                else:
                    cell_x, cell_y, cell_width, cell_height = x, y, width, height
                
                # Get cell content
                content_array = cell.get("content")
                if isinstance(content_array, list):
                    for item in content_array:
                        if not isinstance(item, dict):
                                continue
                        
                        layout_payload = item.get("layout_payload") or item.get("_layout_payload")
                        if layout_payload and isinstance(layout_payload, dict) and layout_payload.get("lines"):
                            # Render as paragraph
                            self._render_paragraph_with_lines(
                                cell_x,
                                cell_y,
                                cell_width,
                                cell_height,
                                item,
                            )
                        else:
                            # Try to render as simple element
                            item_type = item.get("type", "paragraph")
                            match item_type:
                                case "paragraph":
                                    text = item.get("text") or ""
                                    runs = item.get("runs", [])
                                    if runs and isinstance(runs, list):
                                        text_parts = []
                                        for run in runs:
                                            if isinstance(run, dict):
                                                run_text = run.get("text", "")
                                                if run_text:
                                                    text_parts.append(run_text)
                                        text = "".join(text_parts)
                                    
                                    if text:
                                        # TODO: direct Rust draw_string in cell
                                        logger.debug(
                                            f"Would draw text in cell at ({cell_x:.2f}, {cell_y:.2f}): "
                                            f"'{text[:50]}...'"
                                        )
                                
                                case "image":
                                    self._render_image_direct(
                                        cell_x,
                                        cell_y,
                                        cell_width,
                                        cell_height,
                                        item,
                                    )
                                
                                case _:
                                    logger.debug(f"Unknown item type in cell: {item_type}, skipping")
                
                # TODO: Render cell borders via renderer.draw_rect / rectangle if exposed

    # ======================================================================
    # Image rendering
    # ======================================================================

    def _render_image_direct(
        self,
        x: float,
        y: float,
        width: float,
        height: float,
        content: Dict[str, Any],
    ) -> None:
        """
        Render image by calling Rust render_image_block or render_image_block_stream
        with x, y, width, height.
        """
        # Extract image data
        image_data = content.get("data") if isinstance(content.get("data"), dict) else content
        image_path = content.get("path") or content.get("image_path") or ""
        if isinstance(image_data, dict) and not image_path:
            image_path = image_data.get("path") or image_data.get("image_path") or ""
        
        resolved_path = self._resolve_image_path(image_data or content)
        if resolved_path:
            image_path = resolved_path
        
        stream_key = None
        if isinstance(image_data, dict):
            stream_key = image_data.get("stream_key")
            registered = self._register_image_stream(image_data, resolved_path)
            if registered:
                stream_key = registered
        elif isinstance(content, dict):
            stream_key = content.get("stream_key")
            registered = self._register_image_stream(content, resolved_path)
            if registered:
                stream_key = registered
        
        width_emu = content.get("width_emu") or content.get("width")
        height_emu = content.get("height_emu") or content.get("height")
        numeric_width = width_emu if isinstance(width_emu, (int, float)) else None
        numeric_height = height_emu if isinstance(height_emu, (int, float)) else None
        
        if stream_key:
            self.renderer.render_image_block_stream(
                x,
                y,
                width,
                height,
                str(stream_key),
                numeric_width,
                numeric_height,
            )
        elif image_path:
            self.renderer.render_image_block(
                x,
                y,
                width,
                height,
                str(image_path),
                numeric_width,
                numeric_height,
            )
        else:
            logger.warning("Image block missing path and stream_key")

    # ======================================================================
    # Overlay rendering
    # ======================================================================

    def _render_overlay(self, overlay: Dict[str, Any]) -> None:
        """
        Render overlay (floating image/textbox) by calling Rust methods with
        x, y, width, height.
        """
        frame = overlay.get("frame", {})
        if not isinstance(frame, dict):
            return
        
        overlay_x = frame.get("x", 0.0)
        overlay_y = frame.get("y", 0.0)
        overlay_width = frame.get("width", 0.0)
        overlay_height = frame.get("height", 0.0)
        
        overlay_kind = overlay.get("kind", "image")
        payload = overlay.get("payload", {})
        
        match overlay_kind:
            case "image":
                image_data = payload.get("image") if isinstance(payload, dict) else payload
                if isinstance(image_data, dict):
                    stream_key = image_data.get("stream_key")
                    image_path = image_data.get("path") or image_data.get("image_path")
                    if stream_key or image_path:
                        # TODO: Replace with Rust overlay draw_image
                        logger.debug(
                            f"Would draw overlay image at ({overlay_x:.2f}, {overlay_y:.2f}) "
                            f"size ({overlay_width:.2f}, {overlay_height:.2f})"
                        )
            
            case "textbox":
                textbox_content = payload.get("content") if isinstance(payload, dict) else payload
                if isinstance(textbox_content, dict):
                    layout_payload = (
                        textbox_content.get("layout_payload")
                        or textbox_content.get("_layout_payload")
                    )
                    if layout_payload and isinstance(layout_payload, dict) and layout_payload.get("lines"):
                        self._render_paragraph_with_lines(
                            overlay_x,
                            overlay_y,
                            overlay_width,
                            overlay_height,
                            textbox_content,
                        )
            
            case _:
                logger.debug(f"Unknown overlay kind: {overlay_kind}, skipping")

    # ======================================================================
    # Helpers
    # ======================================================================

    def _prepare_content_for_renderer(self, content: Any) -> Dict[str, Any]:
        """
        Prepare content for renderer by converting to dict and ensuring
        layout_payload exists if needed.
        Safely handles circular references by using visited set.
        """
        if isinstance(content, dict):
            return content
        
        # Use safe conversion that handles circular references
        return self._safe_to_dict(content, visited=None)
    
    def _safe_to_dict(self, obj: Any, visited: Optional[Set[int]] = None, depth: int = 0) -> Dict[str, Any]:
        """
        Safely convert object to dict, handling circular references.
        """
        # Prevent infinite recursion
        if depth > 10:
            return {"_error": "max_depth_exceeded", "_type": type(obj).__name__}
        
        if visited is None:
            visited = set()
        
        obj_id = id(obj)
        if obj_id in visited:
            return {"_circular_ref": type(obj).__name__}
        
        visited.add(obj_id)
        
        try:
            if isinstance(obj, dict):
                result = {}
                for k, v in obj.items():
                    if isinstance(k, str) and not k.startswith("__"):
                        result[k] = self._safe_to_dict(v, visited, depth + 1)
                return result
            
            elif isinstance(obj, (list, tuple)):
                return [self._safe_to_dict(item, visited, depth + 1) for item in obj]
            
            elif is_dataclass(obj):
                result = {}
                for field in fields(obj):
                    field_name = field.name
                    if field_name.startswith("_"):
                            continue
                    try:
                        field_value = getattr(obj, field_name, None)
                        if field_value is not None:
                            result[field_name] = self._safe_to_dict(field_value, visited, depth + 1)
                    except Exception:
                        continue
                return result
            
            elif hasattr(obj, "to_dict"):
                try:
                    dict_result = obj.to_dict()
                    if isinstance(dict_result, dict):
                        return self._safe_to_dict(dict_result, visited, depth + 1)
                    return dict_result
                except Exception:
                    pass
            
            elif hasattr(obj, "__dict__"):
                result = {}
                for k, v in obj.__dict__.items():
                    if not k.startswith("__"):
                        try:
                            result[k] = self._safe_to_dict(v, visited, depth + 1)
                        except Exception:
                            continue
                return result
            
            elif isinstance(obj, (str, int, float, bool, type(None))):
                return obj
            
            else:
                # Fallback: return string representation
                return {"_type": type(obj).__name__, "_repr": str(obj)[:100]}
        
        finally:
            visited.discard(obj_id)
    
    def _style_to_dict(self, style: Any) -> Dict[str, Any]:
        """Convert style to dict."""
        if isinstance(style, dict):
            return style
        elif is_dataclass(style):
            return asdict(style)
        elif hasattr(style, "__dict__"):
            return {k: v for k, v in style.__dict__.items() if not k.startswith("__")}
        else:
            return {}
    
    def _resolve_image_path(self, image: Any) -> Optional[str]:
        """Resolve image path from image object/dict."""
        if isinstance(image, dict):
            return image.get("path") or image.get("src") or image.get("image_path")
        elif hasattr(image, "path"):
            return getattr(image, "path", None) or getattr(image, "image_path", None)
        return None
    
    def _register_image_stream(
        self,
        image_data: Dict[str, Any],
        resolved_path: Optional[str] = None,
    ) -> Optional[str]:
        """
        Register image as stream and return stream_key.
        Returns None if image cannot be registered.
        """
        # Check if already registered
        stream_key = image_data.get("stream_key")
        if stream_key and stream_key in self.image_streams:
            return stream_key
        
        # Load image bytes
        image_bytes, mime_type = self._load_image_bytes_for_stream(image_data, resolved_path)
        if not image_bytes:
            return None
        
        # Generate stream_key
        if not stream_key:
            if resolved_path:
                stream_key = f"img_{hashlib.md5(resolved_path.encode()).hexdigest()[:8]}"
            else:
                stream_key = f"img_{hashlib.md5(image_bytes).hexdigest()[:8]}"
        
        # Store in image_streams
        self.image_streams[stream_key] = image_bytes
        
        # Register with Rust renderer
        try:
            self.renderer.register_image_stream(stream_key, image_bytes, mime_type)
            self._registered_stream_keys.add(stream_key)
            return stream_key
        except Exception as e:
            logger.warning(f"Failed to register image stream: {e}")
            return None
    
    def _load_image_bytes_for_stream(
        self,
        image_data: Dict[str, Any],
        resolved_path: Optional[str] = None,
    ) -> Tuple[Optional[bytes], Optional[str]]:
        """
        Load image bytes for streaming to Rust.
        Returns (bytes, mime_type) or (None, None) if failed.
        For WMF/EMF, returns raw bytes with mime_type=None so Rust can handle conversion.
        """
        image_path = resolved_path or image_data.get("path") or image_data.get("image_path")
        
        # Try to load from package_reader
        if self.package_reader and image_path:
            try:
                raw_data = self.package_reader.get_binary_content(image_path)
                if raw_data:
                    # WMF/EMF detection (let Rust convert)
                    if len(raw_data) > 0:
                        if raw_data[:2] == b"\x01\x00" or raw_data[:4] == b"\x01\x00\x00\x00":
                            return raw_data, None
                        mime_type, _ = mimetypes.guess_type(image_path)
                        return raw_data, mime_type
            except Exception as e:
                logger.debug(f"Failed to load image from package_reader: {e}")
        
        # Try filesystem
        if image_path and Path(image_path).exists():
            try:
                raw_data = Path(image_path).read_bytes()
                mime_type, _ = mimetypes.guess_type(image_path)
                return raw_data, mime_type
            except Exception as e:
                logger.debug(f"Failed to load image from filesystem: {e}")

        return None, None
